from reikna.cluda.api_discovery import cuda_id, ocl_id, api_ids, supports_api, \
    supported_api_ids, get_api, cuda_api, ocl_api, any_api
from reikna.cluda.api_tools import find_devices
from reikna.cluda.kernel import Module, Snippet


class OutOfResourcesError(Exception):
    """
    Thrown by :py:meth:`~reikna.cluda.api.Thread.compile_static`
    if the provided ``local_size`` is too big, or one cannot be found.
    """
    pass



import numpy
import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes


def pytest_generate_tests(metafunc):

    if 'dtype_to_align' in metafunc.funcargnames:

        vals = []
        ids = []

        # numpy itemsize == 9, but on device it will be aligned to 4, so the total size will be 12
        dtype = numpy.dtype([('val1', numpy.int32), ('val2', numpy.int32), ('pad', numpy.int8)])
        vals.append(dtype)
        ids.append("small_pad")

        dtype_nested = numpy.dtype([
            ('val1', numpy.int32), ('pad', numpy.int8)])
        dtype = numpy.dtype([
            ('val1', numpy.int32),
            ('val2', numpy.int16),
            ('nested', dtype_nested)])
        vals.append(dtype)
        ids.append("nested")

        dtype_nested = numpy.dtype(dict(
            names=['val1', 'pad'],
            formats=[numpy.int8, numpy.int8]))
        dtype = numpy.dtype(dict(
            names=['pad', 'struct_arr', 'regular_arr'],
            formats=[
                numpy.int32,
                numpy.dtype((dtype_nested, 2)),
                numpy.dtype((numpy.int16, (2, 3)))]))
        vals.append(dtype)
        ids.append("nested_array")

        metafunc.parametrize('dtype_to_align', vals, ids=ids)


def get_offsets_from_device(thr, dtype):
    """
    Returns a new data type object with the same fields as in ``dtype``
    and the field offsets set by the compiler in ``thr``.
    All existing offsets in ``dtype`` are ignored.
    """

    if dtype.names is None:
        return dtype

    struct = dtypes.ctype_module(dtype, ignore_alignment=True)

    program = thr.compile(
    """
    #define my_offsetof(type, field) ((size_t)(&((type *)0)->field))

    KERNEL void test(GLOBAL_MEM int *dest)
    {
      %for i, name in enumerate(dtype.names):
      dest[${i}] = my_offsetof(${struct}, ${name});
      %endfor
      dest[${len(dtype.names)}] = sizeof(${struct});
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=dtype))

    offsets = thr.array(len(dtype.names) + 1, numpy.int32)
    test = program.test
    test(offsets, global_size=1)
    offsets = offsets.get()

    # Casting to Python ints, becase numpy ints as dtype offsets make it unhashable.
    offsets = [int(offset) for offset in offsets]

    aligned_dtypes = [
        get_offsets_from_device(thr, dtype.fields[name][0])
        for name in dtype.names]

    return numpy.dtype(dict(
        names=dtype.names,
        formats=aligned_dtypes,
        offsets=offsets[:-1],
        itemsize=offsets[-1],
        aligned=True))


def test_align(thr, dtype_to_align):
    aligned_dtype = dtypes.align(dtype_to_align)
    empyric_dtype = get_offsets_from_device(thr, dtype_to_align)
    assert aligned_dtype == empyric_dtype


def check_struct_fill(thr, dtype):
    """
    Fill every field of the given ``dtype`` with its number and check the results.
    This helps to detect issues with offsets in the struct.
    """
    struct = dtypes.ctype_module(dtype)

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM ${struct} *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct} res;

      %for i, field_info in enumerate(dtypes.flatten_dtype(dtype)):
      res.${".".join(field_info[0])} = ${i};
      %endfor

      dest[i] = res;
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=dtype))

    test = program.test

    a_dev = thr.array(128, dtype)
    test(a_dev, global_size=128)
    a = a_dev.get()

    for i, field_info in enumerate(dtypes.flatten_dtype(dtype)):
        path, _ = field_info
        assert (dtypes.extract_field(a, path) == i).all()


def test_hardcoded_offsets(thr):
    """
    Test the correctness of alignment for an explicit set of field offsets.
    """

    dtype_nested = numpy.dtype(dict(
        names=['val1', 'pad'],
        formats=[numpy.int32, numpy.int8],
        offsets=[0, 4],
        itemsize=8,
        aligned=True))

    dtype = numpy.dtype(dict(
        names=['val1', 'val2', 'nested'],
        formats=[numpy.int32, numpy.int16, dtype_nested],
        offsets=[0, 4, 8],
        itemsize=32,
        aligned=True))

    check_struct_fill(thr, dtype)


def test_align(thr):
    """
    Test the correctness of alignment for field offsets adjusted automatically.
    """

    dtype_nested = numpy.dtype([
        ('val1', numpy.int32), ('pad', numpy.int8)])

    dtype = numpy.dtype([
        ('val1', numpy.int32),
        ('val2', numpy.int16),
        ('nested', dtype_nested)])

    dtype = dtypes.align(dtype)

    check_struct_fill(thr, dtype)


def test_nested_array(thr):
    """
    Check that structures with nested arrays are processed correctly.
    """
    dtype_nested = numpy.dtype(dict(
        names=['val1', 'pad'],
        formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(dict(
        names=['pad', 'struct_arr', 'regular_arr'],
        formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))]))

    dtype = dtypes.align(dtype)
    struct = dtypes.ctype_module(dtype)

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM ${struct} *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct} res;

      res.struct_arr[0].val1 = i + 0;
      res.struct_arr[1].val1 = i + 1;

      res.regular_arr[0] = i + 2;
      res.regular_arr[1] = i + 3;
      res.regular_arr[2] = i + 4;

      dest[i] = res;
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=dtype))

    test = program.test

    a_dev = thr.array(64, dtype)
    test(a_dev, global_size=64)
    a = a_dev.get()

    idxs = numpy.arange(64)
    assert (a['struct_arr'][:,0]['val1'] == idxs + 0).all()
    assert (a['struct_arr'][:,1]['val1'] == idxs + 1).all()
    assert (a['regular_arr'][:,0] == idxs + 2).all()
    assert (a['regular_arr'][:,1] == idxs + 3).all()
    assert (a['regular_arr'][:,2] == idxs + 4).all()


def test_structural_typing(some_thr):
    """
    Checks that ``ctype_module`` for equal dtype objects result in the same module object
    (which means that these two types will actually be rendered as a single type).
    """
    dtype = dtypes.align(numpy.dtype([('val1', numpy.int32), ('val2', numpy.float32)]))

    struct1 = dtypes.ctype_module(dtype)
    struct2 = dtypes.ctype_module(dtype)

    # Check that these reference the same object
    assert struct1 is struct2

    # Just in case, compile a test kernel
    program = some_thr.compile(
    """
    KERNEL void test(GLOBAL_MEM float *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct1} temp1;
      ${struct2} temp2;

      temp1.val1 = 0;
      temp1.val2 = 1;

      // If struct1 and struct2 correspond to different types,
      // this will give a compilation error,
      // because C has a nominative typing system.
      temp2 = temp1;

      dest[i] = temp2.val1 + temp2.val2;
    }
    """, render_kwds=dict(
        struct1=struct1,
        struct2=struct2))


def test_structural_typing_nested(some_thr):
    """
    Check that the structural typing behavior works for nested structures.
    In other words, a nested dtype gets represented by the same module as
    an equal top-level dtype.
    """
    dtype_nested = dtypes.align(
        numpy.dtype([('val1', numpy.int32), ('val2', numpy.float32)]))
    dtype = dtypes.align(numpy.dtype([
        ('val1', numpy.int32), ('val2', numpy.float32),
        ('nested', dtype_nested)]))

    struct_nested = dtypes.ctype_module(dtype_nested)
    struct = dtypes.ctype_module(dtype)

    program = some_thr.compile(
    """
    KERNEL void test(GLOBAL_MEM float *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct_nested} temp_nested;
      ${struct} temp;

      temp_nested.val1 = 0;
      temp_nested.val2 = 1;

      // If the nested structure has a different type from temp_nested,
      // this will give a compilation error.
      temp.nested = temp_nested;
      temp.val1 = 0;
      temp.val2 = 1;

      dest[i] = temp.val1 + temp.val2 + temp.nested.val1 + temp.nested.val2;
    }
    """, render_kwds=dict(
        struct=struct,
        struct_nested=struct_nested))



"""
General purpose algorithms.


Pure parallel computations
^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: PureParallel
    :members:


Transposition (permutation)
^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: Transpose
    :members:


Reduction
^^^^^^^^^

.. autoclass:: Predicate
    :members:

.. autofunction:: predicate_sum

.. autoclass:: Reduce
    :members:
"""

from reikna.algorithms.pureparallel import PureParallel
from reikna.algorithms.transpose import Transpose
from reikna.algorithms.reduce import Reduce, Predicate, predicate_sum



import numpy
from numpy.lib.stride_tricks import as_strided

from reikna.helpers import wrap_in_tuple
from reikna.cluda import dtypes


# Default tolerances for numpy.allclose().
# Should be enough to detect a error, but not enough to trigger a fail
# in case of a slightly imprecise implementation of some operation on a GPU.
SINGLE_RTOL = 1e-5
SINGLE_ATOL = 1e-8

DOUBLE_RTOL = 1e-11
DOUBLE_ATOL = 1e-11


def get_test_array_like(arr, **kwds):
    kwds['strides'] = arr.strides
    return get_test_array(arr.shape, arr.dtype, **kwds)

def get_test_array(shape, dtype, strides=None, no_zeros=False, high=None):
    shape = wrap_in_tuple(shape)
    dtype = dtypes.normalize_type(dtype)

    if dtype.names is not None:
        result = numpy.empty(shape, dtype)
        for name in dtype.names:
            result[name] = get_test_array(shape, dtype[name], no_zeros=no_zeros, high=high)
    else:
        if dtypes.is_integer(dtype):
            low = 1 if no_zeros else 0
            if high is None:
                high = 100 # will work even with signed chars
            get_arr = lambda: numpy.random.randint(low, high, shape).astype(dtype)
        else:
            low = 0.01 if no_zeros else 0
            if high is None:
                high = 1.0
            get_arr = lambda: numpy.random.uniform(low, high, shape).astype(dtype)

        if dtypes.is_complex(dtype):
            result = get_arr() + 1j * get_arr()
        else:
            result = get_arr()

    if strides is not None:
        result = as_strided(result, result.shape, strides)

    return result


def diff_is_negligible(m, m_ref, atol=None, rtol=None):

    if m.dtype.names is not None:
        return all(diff_is_negligible(m[name], m_ref[name]) for name in m.dtype.names)

    assert m.dtype == m_ref.dtype

    if dtypes.is_integer(m.dtype):
        close = (m == m_ref)
    else:
        if atol is None:
            atol = DOUBLE_ATOL if dtypes.is_double(m.dtype) else SINGLE_ATOL
        if rtol is None:
            rtol = DOUBLE_RTOL if dtypes.is_double(m.dtype) else SINGLE_RTOL

        close = numpy.isclose(m, m_ref, atol=atol, rtol=rtol)

    if close.all():
        return True

    far_idxs = numpy.vstack(numpy.where(close == False)).T
    print(
        ("diff_is_negligible() with atol={atol} and rtol={rtol} " +
        "found {diffs} differences, first ones are:").format(
        atol=atol, rtol=rtol, diffs=str(far_idxs.shape[0])))
    for idx, _ in zip(far_idxs, range(10)):
        idx = tuple(idx)
        print("idx: {idx}, test: {test}, ref: {ref}".format(
            idx=idx, test=m[idx], ref=m_ref[idx]))

    return False



import numpy

from reikna.helpers import template_from, min_blocks
import reikna.cluda.functions as functions
from reikna.core import Computation, Parameter, Annotation, Transformation


# Output = Input * Parameter
def tr_scale(arr, coeff_t):
    return Transformation(
        [Parameter('o1', Annotation(arr, 'o')),
        Parameter('i1', Annotation(arr, 'i')),
        Parameter('s1', Annotation(coeff_t))],
        "${o1.store_same}(${mul}(${i1.load_same}, ${s1}));",
        render_kwds=dict(
            mul=functions.mul(arr.dtype, coeff_t, out_dtype=arr.dtype)))


class Dummy(Computation):
    """
    Dummy computation class with two inputs, two outputs and one parameter.
    Used to perform core and transformation tests.
    """

    def __init__(self, arr1, arr2, coeff, same_A_B=False,
            test_incorrect_parameter_name=False,
            test_untyped_scalar=False,
            test_kernel_adhoc_array=False):

        assert len(arr1.shape) == 2
        assert len(arr2.shape) == (2 if same_A_B else 1)
        assert arr1.dtype == arr2.dtype
        if same_A_B:
            assert arr1.shape == arr2.shape
        else:
            assert arr1.shape[0] == arr1.shape[1]

        self._same_A_B = same_A_B
        self._persistent_array = numpy.arange(arr2.size).reshape(arr2.shape).astype(arr2.dtype)

        self._test_untyped_scalar = test_untyped_scalar
        self._test_kernel_adhoc_array = test_kernel_adhoc_array

        Computation.__init__(self, [
            Parameter(('_C' if test_incorrect_parameter_name else 'C'), Annotation(arr1, 'o')),
            Parameter('D', Annotation(arr2, 'o')),
            Parameter('A', Annotation(arr1, 'i')),
            Parameter('B', Annotation(arr2, 'i')),
            Parameter('coeff', Annotation(coeff))])

    def _build_plan(self, plan_factory, device_params, C, D, A, B, coeff):
        plan = plan_factory()

        arr_dtype = C.dtype
        coeff_dtype = coeff.dtype

        mul = functions.mul(arr_dtype, coeff_dtype)
        div = functions.div(arr_dtype, coeff_dtype)

        template = template_from("""
        <%def name="dummy(kernel_declaration, C, D, A, B, coeff)">
        ${kernel_declaration}
        {
            VIRTUAL_SKIP_THREADS;
            VSIZE_T idx0 = virtual_global_id(0);
            VSIZE_T idx1 = virtual_global_id(1);

            ${A.ctype} a = ${A.load_idx}(idx0, idx1);
            ${C.ctype} c = ${mul}(a, ${coeff});
            ${C.store_idx}(idx1, idx0, c);

            %if same_A_B:
                ${B.ctype} b = ${B.load_idx}(idx0, idx1);
                ${D.ctype} d = ${div}(b, ${coeff});
                ${D.store_idx}(idx0, idx1, d);
            %else:
            if (idx1 == 0)
            {
                ${B.ctype} b = ${B.load_idx}(idx0);
                ${D.ctype} d = ${div}(b, ${coeff});
                ${D.store_idx}(idx0, d);
            }
            %endif
        }
        </%def>

        <%def name="dummy2(kernel_declaration, CC, DD, C, D, pers_arr, const_coeff)">
        ${kernel_declaration}
        {
            VIRTUAL_SKIP_THREADS;
            VSIZE_T idx0 = virtual_global_id(0);
            VSIZE_T idx1 = virtual_global_id(1);

            ${CC.store_idx}(idx0, idx1, ${C.load_idx}(idx0, idx1));

            %if same_A_B:
                ${DD.store_idx}(
                    idx0, idx1,
                    ${mul}(${D.load_idx}(idx0, idx1), ${const_coeff}) +
                        ${pers_arr.load_idx}(idx0, idx1));
            %else:
            if (idx1 == 0)
            {
                ${DD.store_idx}(
                    idx0,
                    ${mul}(${D.load_idx}(idx0), ${const_coeff}) +
                        ${pers_arr.load_idx}(idx0));
            }
            %endif
        }
        </%def>
        """)

        block_size = 8

        C_temp = plan.temp_array_like(C)
        D_temp = plan.temp_array_like(D)
        arr = plan.persistent_array(self._persistent_array)

        plan.kernel_call(
            template.get_def('dummy'),
            [C_temp, D_temp, A, B, coeff],
            global_size=A.shape,
            local_size=(block_size, block_size),
            render_kwds=dict(mul=mul, div=div, same_A_B=self._same_A_B))

        plan.kernel_call(
            template.get_def('dummy2'),
            [C, D, C_temp, D_temp,
                (self._persistent_array if self._test_kernel_adhoc_array else arr),
                (10 if self._test_untyped_scalar else numpy.float32(10))],
            global_size=A.shape,
            local_size=(block_size, block_size),
            render_kwds=dict(mul=mul, same_A_B=self._same_A_B))

        return plan


# A function which does the same job as base Dummy kernel
def mock_dummy(a, b, coeff):
    return a.T * coeff, (b / coeff * 10) + numpy.arange(b.size).reshape(b.shape).astype(b.dtype)


class DummyNested(Computation):
    """
    Dummy computation class with a nested computation inside.
    """

    def __init__(self, arr1, arr2, coeff, second_coeff, same_A_B=False,
            test_computation_adhoc_array=False,
            test_computation_incorrect_role=False,
            test_computation_incorrect_type=False,
            test_same_arg_as_i_and_o=False):

        self._second_coeff = second_coeff
        self._same_A_B = same_A_B
        self._test_same_arg_as_i_and_o = test_same_arg_as_i_and_o

        self._test_computation_adhoc_array = test_computation_adhoc_array
        self._test_computation_incorrect_role = test_computation_incorrect_role
        self._test_computation_incorrect_type = test_computation_incorrect_type

        Computation.__init__(self, [
            Parameter('C', Annotation(arr1, 'o')),
            Parameter('D', Annotation(arr2, 'o')),
            Parameter('A', Annotation(arr1, 'i')),
            Parameter('B', Annotation(arr2, 'i')),
            Parameter('coeff', Annotation(coeff))])

    def _build_plan(self, plan_factory, device_params, C, D, A, B, coeff):
        plan = plan_factory()
        nested = Dummy(A, B, coeff, same_A_B=self._same_A_B)

        scale = tr_scale(A, numpy.float32)
        nested.parameter.A.connect(scale, scale.o1, A_prime=scale.i1, scale_coeff=scale.s1)

        C_temp = plan.temp_array_like(C)
        D_temp = plan.temp_array_like(D)

        plan.computation_call(
            nested,
            (numpy.empty_like(C_temp) if self._test_computation_adhoc_array else C_temp),
            (B if self._test_computation_incorrect_role else D_temp),
            (B if self._test_computation_incorrect_type else A),
            0.4, B, coeff)

        if self._test_same_arg_as_i_and_o:
            _trash = plan.temp_array_like(C) # ignoring this result
            nested2 = Dummy(A, B, coeff, same_A_B=self._same_A_B)
            plan.computation_call(nested2, _trash, D_temp, C_temp, D_temp, coeff)

        plan.computation_call(nested, C, D, C_temp, 0.4, D_temp, self._second_coeff)

        return plan


def mock_dummy_nested(a, b, coeff, second_coeff, test_same_arg_as_i_and_o=False):
    c, d = mock_dummy(0.4 * a, b, coeff)
    if test_same_arg_as_i_and_o:
        _, d = mock_dummy(c, d, coeff)
    return mock_dummy(0.4 * c, d, second_coeff)


class DummyAdvanced(Computation):
    """
    Dummy computation class which uses some advanced features.
    """

    def __init__(self, arr, coeff):
        Computation.__init__(self, [
            Parameter('C', Annotation(arr, 'io')),
            Parameter('D', Annotation(arr, 'io')),
            Parameter('coeff1', Annotation(coeff)),
            Parameter('coeff2', Annotation(coeff))])

    def _build_plan(self, plan_factory, device_params, C, D, coeff1, coeff2):
        plan = plan_factory()
        nested = Dummy(C, D, coeff1, same_A_B=True)

        C_temp = plan.temp_array_like(C)
        D_temp = plan.temp_array_like(D)

        # Testing a computation call which uses the same argument for two parameters.
        plan.computation_call(nested, C_temp, D, C, C, coeff1)

        arr_dtype = C.dtype
        coeff_dtype = coeff2.dtype

        mul = functions.mul(arr_dtype, coeff_dtype)
        div = functions.div(arr_dtype, coeff_dtype)

        template = template_from("""
        <%def name="dummy(kernel_declaration, CC, C, D, coeff)">
        ${kernel_declaration}
        {
            VIRTUAL_SKIP_THREADS;
            VSIZE_T idx0 = virtual_global_id(0);
            VSIZE_T idx1 = virtual_global_id(1);

            ${CC.store_idx}(idx0, idx1,
                ${C.load_idx}(idx0, idx1) +
                ${mul}(${D.load_idx}(idx0, idx1), ${coeff}));
        }
        </%def>
        """)

        # Testing a kernel call which uses the same argument for two parameters.
        plan.kernel_call(
            template.get_def('dummy'),
            [C, C_temp, C_temp, coeff2],
            global_size=C.shape,
            render_kwds=dict(mul=mul))

        return plan


def mock_dummy_advanced(c, d, coeff1, coeff2):
    ct, d = mock_dummy(c, c, coeff1)
    c = ct + coeff2 * ct
    return c, d



"""
Fast Fourier Transform
^^^^^^^^^^^^^^^^^^^^^^

.. autoclass :: FFT
    :members:

FFT frequency shift
^^^^^^^^^^^^^^^^^^^

.. autoclass :: FFTShift
    :members:
"""

from reikna.fft.fft import FFT
from reikna.fft.fftshift import FFTShift



import time
import numpy
import pytest

from helpers import *

from reikna.linalg import MatrixMul
import reikna.cluda.dtypes as dtypes
from reikna.cluda import OutOfResourcesError
from reikna.helpers import product


def pytest_generate_tests(metafunc):

    if 'perf_bwo' in metafunc.funcargnames:
        bwos = [8, 16, 32]
        ids=["8x8", "16x16", "32x32"]
        metafunc.parametrize('perf_bwo', bwos, ids=ids)

    if 'perf_shape' in metafunc.funcargnames:

        mem_limit = 2 ** 20
        sizes = [16, 32, 64, 256, 512, 25]

        perf_shapes = []

        for size in sizes:
            perf_shapes.append((mem_limit // size ** 2, size))

        ids = []
        for batch, size in perf_shapes:
            ids.append(str(batch) + 'x' + str(size) + "," + str(size))

        metafunc.parametrize('perf_shape', perf_shapes, ids=ids)

    if 'sizes' in metafunc.funcargnames:
        sizes = [
            (15, 17, 3), (122, 5, 1000), (45, 99, 40), (56, 78, 44)]
        ids = [str(size) for size in sizes]
        metafunc.parametrize('sizes', sizes, ids=ids)

    if 'batches' in metafunc.funcargnames:
        batches = [
            (tuple(), tuple()),
            (tuple(), (14,)),
            ((35,), tuple()),
            ((12,), (12,))]
        ids = [str(batch) for batch in batches]
        metafunc.parametrize('batches', batches, ids=ids)

    if 'transposed_a' in metafunc.funcargnames:
        metafunc.parametrize('transposed_a', [False, True], ids=['A', 'A.T'])

    if 'transposed_b' in metafunc.funcargnames:
        metafunc.parametrize('transposed_b', [False, True], ids=['B', 'B.T'])

    if 'arg_dtypes' in metafunc.funcargnames:
        arg_dtypes = [(False, False), (False, True), (True, False), (True, True)]
        mark = lambda x: 'c' if x else 'r'
        ids = [mark(t1) + mark(t2) for t1, t2 in arg_dtypes]
        metafunc.parametrize('arg_dtypes', arg_dtypes, ids=ids)


def ref_dot(a, b):
    a_batch = product(a.shape[:-2])
    b_batch = product(b.shape[:-2])

    assert a_batch == b_batch or a_batch == 1 or b_batch == 1

    a = a.reshape(a_batch, a.shape[-2], a.shape[-1])
    b = b.reshape(b_batch, b.shape[-2], b.shape[-1])

    out_batch = max(a_batch, b_batch)
    out_shape = (out_batch, a.shape[-2], b.shape[-1])
    out_dtype = numpy.result_type(a.dtype, b.dtype)
    out = numpy.empty(out_shape, out_dtype)

    for i in range(out_batch):
        ai = 0 if a_batch == 1 else i
        bi = 0 if b_batch == 1 else i
        out[i] = numpy.dot(a[ai], b[bi])

    if a_batch == b_batch == 1:
        out = out.reshape(out.shape[-2], out_shape[-1])

    return out


def transpose(m):
    axes = list(range(len(m.shape)))
    axes[-1], axes[-2] = axes[-2], axes[-1]
    return m.transpose(*axes)


def check_errors(thr, a_shape, a_dtype, b_shape, b_dtype, transposed_a=False, transposed_b=False):
    a = get_test_array(a_shape, a_dtype)
    b = get_test_array(b_shape, b_dtype)

    a_ref = transpose(a) if transposed_a else a
    b_ref = transpose(b) if transposed_b else b

    res_ref = ref_dot(a_ref, b_ref)

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    res_dev = thr.empty_like(res_ref)

    dot = MatrixMul(a_dev, b_dev, out_arr=res_dev,
        transposed_a=transposed_a, transposed_b=transposed_b)
    dotc = dot.compile(thr)
    dotc(res_dev, a_dev, b_dev)

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_shapes(thr, batches, sizes):

    a_size, convolution_size, b_size = sizes
    a_batch, b_batch = batches

    a_shape = (a_size, convolution_size)
    b_shape = (convolution_size, b_size)

    check_errors(thr, a_batch + a_shape, numpy.float32, b_batch + b_shape, numpy.float32)


def test_transposed(thr, sizes, transposed_a, transposed_b):

    a_size, convolution_size, b_size = sizes
    a_batch = (10,)
    b_batch = (10,)

    if transposed_a:
        a_shape = (convolution_size, a_size)
    else:
        a_shape = (a_size, convolution_size)

    if transposed_b:
        b_shape = (b_size, convolution_size)
    else:
        b_shape = (convolution_size, b_size)

    check_errors(
        thr, a_batch + a_shape, numpy.float32, b_batch + b_shape, numpy.float32,
        transposed_a=transposed_a, transposed_b=transposed_b)


def test_dtypes(thr_and_double, arg_dtypes):

    thr, double = thr_and_double
    c1, c2 = arg_dtypes

    dtype = numpy.float64 if double else numpy.float32
    dtype1 = dtypes.complex_for(dtype) if c1 else dtype
    dtype2 = dtypes.complex_for(dtype) if c2 else dtype

    check_errors(thr, (30, 40, 50), dtype1, (30, 50, 60), dtype2)


def test_out_arr_shape():
    a = numpy.empty((1, 22, 33), numpy.float32)
    b = numpy.empty((2, 3, 33, 44), numpy.float32)
    dot = MatrixMul(a, b)
    assert dot.parameter.output.shape == (2, 3, 22, 44)


def check_performance(thr_and_double, perf_shape,
        bwo=None, transposed_a=False, transposed_b=False):

    thr, double = thr_and_double
    dtype = numpy.float64 if double else numpy.float32
    batch, size = perf_shape

    shape = (batch, size, size)

    a = get_test_array(shape, dtype)
    b = get_test_array(shape, dtype)

    a_ref = transpose(a) if transposed_a else a
    b_ref = transpose(b) if transposed_b else b

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    res_ref = ref_dot(a_ref, b_ref)
    res_dev = thr.array(res_ref.shape, dtype=dtype)

    dot = MatrixMul(a_dev, b_dev, out_arr=res_dev, block_width_override=bwo,
        transposed_a=transposed_a, transposed_b=transposed_b)

    try:
        dotc = dot.compile(thr)
    except ValueError:
        pytest.skip()

    attempts = 10
    t1 = time.time()
    for i in range(attempts):
        dotc(res_dev, a_dev, b_dev)
    thr.synchronize()
    t2 = time.time()

    assert diff_is_negligible(thr.from_device(res_dev), res_ref)

    return (t2 - t1) / attempts, batch * size ** 3 * 2


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_performance_shape(thr_and_double, perf_shape, transposed_a, transposed_b):
    return check_performance(thr_and_double, perf_shape,
        transposed_a=transposed_a, transposed_b=transposed_b)


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_performance_block_width(thr_and_double, perf_bwo):
    return check_performance(thr_and_double, (4, 512), bwo=perf_bwo)



"""
Linear algebra algorithms.


Matrix multiplication (dot product)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: MatrixMul
    :members:


Matrix norms
^^^^^^^^^^^^

.. autoclass:: EntrywiseNorm
    :members:
"""

from reikna.linalg.matrixmul import MatrixMul
from reikna.linalg.norm import EntrywiseNorm



import numpy

import reikna.helpers as helpers
from reikna.core import Computation, Parameter, Annotation
import reikna.cluda.dtypes as dtypes
from reikna.algorithms import PureParallel
from reikna.transformations import copy

TEMPLATE = helpers.template_for(__file__)


class FFTShift(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Shift the zero-frequency component to the center of the spectrum.
    The interface is similar to ``numpy.fft.fftshift``,
    and the output is the same for the same array shape and axes.

    :param arr_t: an array-like defining the problem array.
    :param axes: a tuple with axes over which to perform the shift.
        If not given, the shift is performed over all the axes.

    .. py:method:: compiled_signature(output:o, input:i)

        ``output`` and ``input`` may be the same array.

        :param output: an array with the attributes of ``arr_t``.
        :param input: an array with the attributes of ``arr_t``.
    """

    def __init__(self, arr_t, axes=None):

        Computation.__init__(self, [
            Parameter('output', Annotation(arr_t, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

        if axes is None:
            axes = tuple(range(len(arr_t.shape)))
        else:
            axes = tuple(axes)
        self._axes = axes

    def _build_trivial_plan(self, plan_factory, output, input_):
        # Trivial problem. Need to add a dummy kernel
        # because we still have to run transformations.

        plan = plan_factory()

        copy_trf = copy(input_, out_arr_t=output)
        copy_comp = PureParallel.from_trf(copy_trf, copy_trf.input)
        plan.computation_call(copy_comp, output, input_)

        return plan

    def _build_plan(self, plan_factory, device_params, output, input_):

        if helpers.product([input_.shape[i] for i in self._axes]) == 1:
            return self._build_trivial_plan(plan_factory, output, input_)

        plan = plan_factory()

        axes = tuple(sorted(self._axes))
        shape = list(input_.shape)

        if all(shape[axis] % 2 == 0 for axis in axes):
        # If all shift axes have even length, it is possible to perform the shift inplace
        # (by swapping pairs of elements).
            shape[axes[0]] //= 2
            plan.kernel_call(
                TEMPLATE.get_def('fftshift_inplace'), [output, input_],
                global_size=shape,
                render_kwds=dict(axes=axes))
        else:
        # Resort to an out-of-place shift to a temporary array and then copy.
            temp = plan.temp_array_like(output)
            plan.kernel_call(
                TEMPLATE.get_def('fftshift_outplace'), [temp, input_],
                global_size=shape,
                render_kwds=dict(axes=axes))

            copy_trf = copy(input_, out_arr_t=output)
            copy_comp = PureParallel.from_trf(copy_trf, copy_trf.input)
            plan.computation_call(copy_comp, output, temp)

        return plan



import sys

import pyopencl as cl
import pyopencl.array as clarray

from reikna.helpers import wrap_in_tuple
import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.api as api_base


def get_id():
    return cluda.ocl_id()


def get_platforms():
    return cl.get_platforms()


class Array(clarray.Array):
    """
    A superclass of PyOpenCL ``Array``, with some additional functionality.
    """
    def __init__(self, thr, *args, **kwds):
        clarray.Array.__init__(self, thr._queue, *args, **kwds)
        self.thread = thr


class Thread(api_base.Thread):

    api = sys.modules[__name__]

    def _process_cqd(self, cqd):
        if isinstance(cqd, cl.Device):
            context = cl.Context(devices=[cqd])
            return context, cl.CommandQueue(context), cqd, False
        elif isinstance(cqd, cl.Context):
            return cqd, cl.CommandQueue(cqd), cqd.devices[0], False
        elif isinstance(cqd, cl.CommandQueue):
            return cqd.context, cqd, cqd.device, False
        else:
            return ValueError("The value provided is not Device, Context or CommandQueue")

    def allocate(self, size):
        return cl.Buffer(self._context, cl.mem_flags.READ_WRITE, size=size)

    def array(self, shape, dtype, strides=None, allocator=None):
        return Array(self, shape, dtype, strides=strides, allocator=allocator)

    def _copy_array(self, dest, src):
        dest.set(src, queue=self._queue, async=self._async)

    def from_device(self, arr, dest=None, async=False):
        arr_cpu = arr.get(queue=self._queue, ary=dest, async=async)
        if dest is None:
            return arr_cpu

    def _copy_array_buffer(self, dest, src, nbytes, src_offset=0, dest_offset=0):
        cl.enqueue_copy(
            self._queue, dest.data, src.data,
            byte_count=nbytes, src_offset=src_offset, dest_offset=dest_offset)

    def synchronize(self):
        self._queue.finish()

    def _compile(self, src, fast_math=False):
        options = "-cl-mad-enable -cl-fast-relaxed-math" if fast_math else ""
        return cl.Program(self._context, src).build(options=options)


class DeviceParameters:

    def __init__(self, device):

        self._device = device

        if device.platform.name == 'Apple' and device.type == cl.device_type.CPU:
        # Apple is being funny again.
        # On OSX 10.8.0 it reports the maximum block size as 1024, when it is really 128.
        # Moreover, if local_barrier() is used in the kernel, it becomes 1
            self.max_work_group_size = 1
            self.max_work_item_sizes = [1, 1, 1]
        else:
            self.max_work_group_size = device.max_work_group_size
            self.max_work_item_sizes = device.max_work_item_sizes

        max_size = 2 ** device.address_bits
        self.max_num_groups = [max_size, max_size, max_size]

        if device.type == cl.device_type.CPU:
            # For CPU both values do not make much sense
            self.local_mem_banks = self.max_work_group_size
            self.warp_size = 1
        elif "cl_nv_device_attribute_query" in device.extensions:
            # If NV extensions are available, use them to query info
            self.local_mem_banks = 16 if device.compute_capability_major_nv < 2 else 32
            self.warp_size = device.warp_size_nv
        elif device.vendor == 'NVIDIA':
            # nVidia device, but no extensions.
            # Must be APPLE OpenCL implementation.
            self.local_mem_banks = 16
            self.warp_size = 16
        else:
            # AMD card.
            # Do not know how to query this info, so settle for most probable values.

            self.local_mem_banks = 32

            # An alternative is to query CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE
            # for some arbitrary kernel.
            self.warp_size = 64

        self.min_mem_coalesce_width = {4: 16, 8: 16, 16: 8}
        self.local_mem_size = device.local_mem_size

    def supports_dtype(self, dtype):
        if dtypes.is_double(dtype):
            extensions = self._device.extensions
            return "cl_khr_fp64" in extensions or "cl_amd_fp64" in extensions
        else:
            return True


class Kernel(api_base.Kernel):

    def _get_kernel(self, program, name):
        return getattr(program, name)

    def _fill_attributes(self):
        self.max_work_group_size = self._kernel.get_work_group_info(
            cl.kernel_work_group_info.WORK_GROUP_SIZE, self._thr._device)

    def prepare(self, global_size, local_size=None, local_mem=0):
        # ``local_mem`` is ignored, since it cannot be easily passed to the kernel
        # (a special kernel argument is requred).
        if local_size is None:
            self._local_size = None
        else:
            self._local_size = wrap_in_tuple(local_size)
        self._global_size = wrap_in_tuple(global_size)

    def _prepared_call(self, *args):
        args = [x.data if isinstance(x, clarray.Array) else x for x in args]
        self._kernel(self._thr._queue, self._global_size, self._local_size, *args)



"""
This module is based on the paper by Salmon et al.,
`P. Int. C. High. Perform. 16 (2011) <http://dx.doi.org/doi:10.1145/2063384.2063405>`_.
and the source code of `Random123 library <http://www.thesalmons.org/john/random123/>`_.

A counter-based random-number generator (CBRNG) is a parametrized function :math:`f_k(c)`,
where :math:`k` is the key, :math:`c` is the counter, and the function :math:`f_k` defines
a bijection in the set of integer numbers.
Being applied to successive counters, the function produces a sequence of pseudo-random numbers.
The key is an analogue of the seed of stateful RNGs;
if the CBRNG is used to generate random num bers in parallel threads, the key is a combination
of a seed and a unique thread number.

There are two types of generators available, ``threefry`` (uses large number of simple functions),
and ``philox`` (uses smaller number of more complicated functions).
The latter one is generally faster on GPUs; see the paper above for detailed comparisons.
These generators can be further specialized to use ``words=2`` or ``words=4``
``bitness=32``-bit or ``bitness=64``-bit counters.
Obviously, the period of the generator equals to the cardinality of the set of possible counters.
For example, if the counter consits of 4 64-bit numbers,
then the period of the generator is :math:`2^{256}`.
As for the key size, in case of ``threefry`` the key has the same size as the counter,
and for ``philox`` the key is half its size.

The :py:class:`~reikna.cbrng.CBRNG` class sets one of the words of the key
(except for ``philox-2x64``, where 32 bit of the only word in the key are used),
the rest are the same for all threads and are derived from the provided ``seed``.
This limits the maximum number of number-generating threads (``size``).
``philox-2x32`` has a 32-bit key and therefore cannot be used in :py:class:`~reikna.cbrng.CBRNG`
(although it can be used separately with the help of the kernel API).

The :py:class:`~reikna.cbrng.CBRNG` class itself is stateless, same as other computations in Reikna,
so you have to manage the generator state yourself.
The state is created by the :py:meth:`~reikna.cbrng.CBRNG.create_counters` method
and contains a ``size`` counters.
This state is then passed to, and updated by a :py:class:`~reikna.cbrng.CBRNG` object.


.. autoclass:: CBRNG
    :members:


Kernel API
^^^^^^^^^^

.. automodule:: reikna.cbrng.bijections
    :members:

.. automodule:: reikna.cbrng.samplers
    :members:

.. automodule:: reikna.cbrng.tools
    :members:
"""

from reikna.cbrng.cbrng import CBRNG



"""
This example demonstrates how to implement a FFT frequency shift (``reikna.fft.FFTShift``)
as a transformation instead of a computation. A peculiarity of this transformation
is the repositioning of elements it performs (as opposed to more common
``load_same``/``store_same`` pair which keel the element order).
It makes this transformation unsafe to use for inplace kernels.

It also contains some performance tests
that compare the speed of FFT + shift as two separate computations and as a
single computation with a transformation against ``numpy`` implementation.
"""

import time
import numpy

from reikna.cluda import any_api
from reikna.fft import FFT, FFTShift

import reikna.cluda.dtypes as dtypes
from reikna.core import Transformation, Parameter, Annotation, Type


def fftshift(arr_t, axes=None):
    """
    Returns a frequency shift transformation (1 output, 1 input) that
    works as ``output = numpy.fft.fftshift(input, axes=axes)``.

    .. warning::

        Involves repositioning of the elements, so cannot be used on inplace kernels.
    """

    if axes is None:
        axes = tuple(range(len(arr_t.shape)))
    else:
        axes = tuple(sorted(axes))

    # The code taken from the FFTShift template for odd problem sizes
    # (at the moment of the writing).
    # Note the use of ``idxs`` template parameter to get access to element indices.
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        """
        <%
            dimensions = len(output.shape)
            new_idx_names = ['new_idx' + str(i) for i in range(dimensions)]
        %>
        %for dim in range(dimensions):
        VSIZE_T ${new_idx_names[dim]} =
            ${idxs[dim]}
            %if dim in axes:
                %if output.shape[dim] % 2 == 0:
                + (${idxs[dim]} < ${output.shape[dim] // 2} ?
                    ${output.shape[dim] // 2} :
                    ${-output.shape[dim] // 2})
                %else:
                + (${idxs[dim]} <= ${output.shape[dim] // 2} ?
                    ${output.shape[dim] // 2} :
                    ${-(output.shape[dim] // 2 + 1)})
                %endif
            %endif
            ;
        %endfor

        ${output.ctype} val = ${input.load_same};
        ${output.store_idx}(${', '.join(new_idx_names)}, val);
        """,
        connectors=['input'],
        render_kwds=dict(axes=axes))


def run_test(thr, shape, dtype, axes=None):

    data = numpy.random.normal(size=shape).astype(dtype)

    fft = FFT(data, axes=axes)
    fftc = fft.compile(thr)

    shift = FFTShift(data, axes=axes)
    shiftc = shift.compile(thr)

    # FFT + shift as two separate computations

    data_dev = thr.to_device(data)

    t_start = time.time()
    fftc(data_dev, data_dev)
    thr.synchronize()
    t_gpu_fft = time.time() - t_start

    t_start = time.time()
    shiftc(data_dev, data_dev)
    thr.synchronize()
    t_gpu_shift = time.time() - t_start

    data_dev = thr.to_device(data)

    t_start = time.time()
    fftc(data_dev, data_dev)
    shiftc(data_dev, data_dev)
    thr.synchronize()
    t_gpu_separate = time.time() - t_start

    data_gpu = data_dev.get()

    # FFT + shift as a computation with a transformation

    data_dev = thr.to_device(data)

    # a separate output array to avoid unsafety of the shift transformation
    res_dev = thr.empty_like(data_dev)

    shift_tr = fftshift(data, axes=axes)
    fft2 = fft.parameter.output.connect(shift_tr, shift_tr.input, new_output=shift_tr.output)
    fft2c = fft2.compile(thr)

    t_start = time.time()
    fft2c(res_dev, data_dev)
    thr.synchronize()
    t_gpu_combined = time.time() - t_start

    # Reference calculation with numpy

    t_start = time.time()
    numpy.fft.fftn(data, axes=axes)
    t_cpu_fft = time.time() - t_start

    t_start = time.time()
    numpy.fft.fftshift(data, axes=axes)
    t_cpu_shift = time.time() - t_start

    t_start = time.time()
    data_ref = numpy.fft.fftn(data, axes=axes)
    data_ref = numpy.fft.fftshift(data_ref, axes=axes)
    t_cpu_all = time.time() - t_start

    data_gpu2 = res_dev.get()

    # Checking that the results are correct
    # (note: this will require relaxing the tolerances
    # if complex64 is used instead of complex128)
    assert numpy.allclose(data_ref, data_gpu)
    assert numpy.allclose(data_ref, data_gpu2)

    return dict(
        t_gpu_fft=t_gpu_fft,
        t_gpu_shift=t_gpu_shift,
        t_gpu_separate=t_gpu_separate,
        t_gpu_combined=t_gpu_combined,
        t_cpu_fft=t_cpu_fft,
        t_cpu_shift=t_cpu_shift,
        t_cpu_all=t_cpu_all)


def run_tests(thr, shape, dtype, axes=None, attempts=10):
    results = [run_test(thr, shape, dtype, axes=axes) for i in range(attempts)]
    return {key:min(result[key] for result in results) for key in results[0]}

if __name__ == '__main__':
    api = any_api()
    thr = api.Thread.create()

    shape = (1024, 1024)
    dtype = numpy.complex128
    axes = (0, 1)

    results = run_tests(thr, shape, dtype, axes=axes)

    print('device:', thr._device.name)
    print('shape:', shape)
    print('dtype:', dtype)
    print('axes:', axes)

    for key, val in results.items():
        print(key, ':', val)

    print(
        "Speedup for a separate calculation:",
        results['t_cpu_all'] / results['t_gpu_separate'])

    print(
        "Speedup for a combined calculation:",
        results['t_cpu_all'] / results['t_gpu_combined'])

    print(
        "Speedup for fft alone:",
        results['t_cpu_fft'] / results['t_gpu_fft'])

    print(
        "Speedup for shift alone:",
        results['t_cpu_shift'] / results['t_gpu_shift'])



import itertools
from fractions import gcd

import numpy
from reikna.helpers import bounding_power_of_2, log2, min_blocks
from reikna.cluda.api_discovery import cuda_id, ocl_id


_DTYPE_TO_BUILTIN_CTYPE = {}
_DTYPE_TO_CTYPE_MODULE = {}


def is_complex(dtype):
    """
    Returns ``True`` if ``dtype`` is complex.
    """
    dtype = normalize_type(dtype)
    return dtype.kind == 'c'

def is_double(dtype):
    """
    Returns ``True`` if ``dtype`` is double precision floating point.
    """
    dtype = normalize_type(dtype)
    return dtype.name in ['float64', 'complex128']

def is_integer(dtype):
    """
    Returns ``True`` if ``dtype`` is an integer.
    """
    dtype = normalize_type(dtype)
    return dtype.kind in ('i', 'u')

def is_real(dtype):
    """
    Returns ``True`` if ``dtype`` is a real.
    """
    dtype = normalize_type(dtype)
    return dtype.kind == 'f'

def _promote_dtype(dtype):
    # not all numpy datatypes are supported by GPU, so we may need to promote
    dtype = normalize_type(dtype)
    if dtype.kind == 'i' and dtype.itemsize < 4:
        dtype = numpy.int32
    elif dtype.kind == 'f' and dtype.itemsize < 4:
        dtype = numpy.float32
    elif dtype.kind == 'c' and dtype.itemsize < 8:
        dtype = numpy.complex64
    return normalize_type(dtype)

def result_type(*dtypes):
    """
    Wrapper for ``numpy.result_type``
    which takes into account types supported by GPUs.
    """
    return _promote_dtype(numpy.result_type(*dtypes))

def min_scalar_type(val):
    """
    Wrapper for ``numpy.min_scalar_dtype``
    which takes into account types supported by GPUs.
    """
    return _promote_dtype(numpy.min_scalar_type(val))

def detect_type(val):
    """
    Find out the data type of ``val``.
    """
    if hasattr(val, 'dtype'):
        return _promote_dtype(val.dtype)
    else:
        return min_scalar_type(val)

def normalize_type(dtype):
    """
    Function for wrapping all dtypes coming from the user.
    ``numpy`` uses two different classes to represent dtypes,
    and one of them does not have some important attributes.
    """
    return numpy.dtype(dtype)

def normalize_types(dtypes):
    """
    Same as :py:func:`normalize_type`, but operates on a list of dtypes.
    """
    return [normalize_type(dtype) for dtype in dtypes]

def complex_for(dtype):
    """
    Returns complex dtype corresponding to given floating point ``dtype``.
    """
    dtype = normalize_type(dtype)
    return numpy.dtype(dict(float32='complex64', float64='complex128')[dtype.name])

def real_for(dtype):
    """
    Returns floating point dtype corresponding to given complex ``dtype``.
    """
    dtype = normalize_type(dtype)
    return numpy.dtype(dict(complex64='float32', complex128='float64')[dtype.name])

def complex_ctr(dtype):
    """
    Returns name of the constructor for the given ``dtype``.
    """
    return 'COMPLEX_CTR(' + ctype(dtype) + ')'

def zero_ctr(dtype):
    """
    Returns the string with constructed zero value for the given ``dtype``.
    """
    if is_complex(dtype):
        return complex_ctr(dtype) + '(0, 0)'
    else:
        return '0'

def cast(dtype):
    """
    Returns function that takes one argument and casts it to ``dtype``.
    """
    def _cast(val):
        # Numpy cannot handle casts to struct dtypes (#4148),
        # so we're avoiding unnecessary casts.
        if not hasattr(val, 'dtype') or val.dtype != dtype:
            return numpy.cast[dtype](val)
        else:
            return val
    return _cast

def _c_constant_arr(val, shape):
    if len(shape) == 0:
        return c_constant(val)
    else:
        return "{" + ", ".join(_c_constant_arr(val[i], shape[1:]) for i in range(shape[0])) + "}"

def c_constant(val, dtype=None):
    """
    Returns a C-style numerical constant.
    If ``val`` has a struct dtype, the generated constant will have the form ``{ ... }``
    and can be used as an initializer for a variable.
    """
    if dtype is None:
        dtype = detect_type(val)
    else:
        dtype = normalize_type(dtype)

    val = cast(dtype)(val)

    if len(val.shape) > 0:
        return _c_constant_arr(val, val.shape)
    elif dtype.names is not None:
        return "{" + ", ".join([c_constant(val[name]) for name in dtype.names]) + "}"

    if is_complex(dtype):
        return "COMPLEX_CTR(" + ctype(dtype) + ")(" + \
            c_constant(val.real) + ", " + c_constant(val.imag) + ")"
    elif is_integer(dtype):
        return str(val) + ("L" if dtype.itemsize > 4 else "")
    else:
        return repr(float(val)) + ("f" if dtype.itemsize <= 4 else "")

def _register_dtype(dtype, ctype_str):
    dtype = normalize_type(dtype)
    _DTYPE_TO_BUILTIN_CTYPE[dtype] = ctype_str

# Taken from compyte.dtypes
def _fill_dtype_registry(respect_windows=True):

    import sys
    import platform

    _register_dtype(numpy.bool, "bool")
    _register_dtype(numpy.int8, "char")
    _register_dtype(numpy.uint8, "unsigned char")
    _register_dtype(numpy.int16, "short")
    _register_dtype(numpy.uint16, "unsigned short")
    _register_dtype(numpy.int32, "int")
    _register_dtype(numpy.uint32, "unsigned int")

    if platform.system == 'Windows' and respect_windows:
        i64_name = "long long"
    else:
        i64_name = "long"

    _register_dtype(numpy.int64, i64_name)
    _register_dtype(numpy.uint64, "unsigned %s" % i64_name)

    _register_dtype(numpy.float32, "float")
    _register_dtype(numpy.float64, "double")
    _register_dtype(numpy.complex64, "float2")
    _register_dtype(numpy.complex128, "double2")

_fill_dtype_registry()


def ctype(dtype):
    """
    For a built-in C type, returns a string with the name of the type.
    """
    return _DTYPE_TO_BUILTIN_CTYPE[normalize_type(dtype)]


def _alignment_str(alignment):
    if alignment is not None:
        return "ALIGN(" + str(alignment) + ")"
    else:
        return ""


def _lcm(*nums):
    """
    Returns the least common multiple of ``nums``.
    """
    if len(nums) == 1:
        return nums[0]
    elif len(nums) == 2:
        return nums[0] * nums[1] // gcd(nums[0], nums[1])
    else:
        return _lcm(nums[0], _lcm(*nums[1:]))


def _struct_alignment(alignments):
    """
    Returns the minimum alignment for a structure given alignments for its fields.
    According to the C standard, it the lowest common multiple of the alignments
    of all of the members of the struct rounded up to the nearest power of two.
    """
    return bounding_power_of_2(_lcm(*alignments))


def _find_minimal_alignment(offset, base_alignment, prev_end):
    """
    Returns the minimal alignment that must be set for a field with
    ``base_alignment`` (the one inherent to the type),
    so that the compiler positioned it at ``offset`` given that the previous field
    ends at the position ``prev_end``.
    """
    # Essentially, we need to find the minimal k such that:
    # 1) offset = m * base_alignment * 2**k, where m > 0 and k >= 0;
    #    (by definition of alignment)
    # 2) offset - prev_offset < base_alignment * 2**k
    #    (otherwise the compiler can just as well take m' = m - 1).
    alignment = base_alignment
    while True:
        if offset % alignment != 0:
            raise ValueError(
                ("Field cannot be positioned at offset {offset}, "
                "since it is not multiple of the minimal alignment {alignment}").format(
                    offset=offset, alignment=alignment))
        if alignment > offset:
            raise ValueError(
                "Could not find a suitable alignment for the field at offset {offset}".format(
                    offset=offset))
        if offset - prev_end >= alignment:
            alignment *= 2
            continue

        return alignment


def _find_alignments(dtype):
    """
    Returns a tuple (base_alignment, field_alignments) for the given dtype, where:

    ``field_alignments`` is a dictionary ``{name:alignment}`` with base alignments
    for every field of ``dtype``.

    ``base_alignment`` is the base alignment for the whole C type corresponding to ``dtype``.
    """

    # FIXME: for vector types with 3 components the alignment is 4*component_size.
    # But we do not support these at the moment anyway.
    if dtype.names is None:
        return dtype.base.itemsize, None

    field_alignments = {}
    explicit_alignments = []
    for i, name in enumerate(dtype.names):
        elem_dtype, elem_offset = dtype.fields[name]
        base_elem_dtype = elem_dtype.base

        base_alignment, _ = _find_alignments(base_elem_dtype)

        if i == 0:
            # The alignment of the first element does not matter, its offset is 0
            explicit_alignments.append(base_alignment)
            field_alignments[name] = None
        else:
            prev_dtype, prev_offset = dtype.fields[dtype.names[i-1]]
            prev_end = prev_offset + prev_dtype.itemsize
            alignment = _find_minimal_alignment(elem_offset, base_alignment, prev_end)

            explicit_alignments.append(alignment)
            field_alignments[name] = alignment if alignment != base_alignment else None

    struct_alignment = _struct_alignment(explicit_alignments)
    last_dtype, last_offset = dtype.fields[dtype.names[-1]]
    last_end = last_offset + last_dtype.itemsize
    struct_alignment = _find_minimal_alignment(dtype.itemsize, struct_alignment, last_end)

    return struct_alignment, field_alignments


def _get_struct_module(dtype, ignore_alignment=False):
    """
    Builds and returns a module with the C type definition for a given ``dtype``,
    possibly using modules for nested structures.
    """
    if not ignore_alignment:
        struct_alignment, field_alignments = _find_alignments(dtype)

    # FIXME: the tag (${prefix}_) is not necessary, but it helps to avoid
    # CUDA bug #1409907 (nested struct initialization like
    # "mystruct x = {0, {0, 0}, 0};" fails to compile)
    lines = ["typedef struct ${prefix}_ {"]
    kwds = {}
    for name in dtype.names:
        elem_dtype, elem_offset = dtype.fields[name]

        if len(elem_dtype.shape) == 0:
            base_elem_dtype = elem_dtype
            elem_dtype_shape = tuple()
        else:
            base_elem_dtype = elem_dtype.base
            elem_dtype_shape = elem_dtype.shape

        if len(elem_dtype_shape) == 0:
            array_suffix = ""
        else:
            array_suffix = "".join("[" + str(d) + "]" for d in elem_dtype_shape)

        typename_var = "typename_" + name
        field_alignment = None if ignore_alignment else field_alignments[name]
        lines.append(
            "    ${" + typename_var + "} " +
            _alignment_str(field_alignment) + " " +
            name + array_suffix + ";")
        kwds[typename_var] = ctype_module(base_elem_dtype, ignore_alignment=ignore_alignment)

    struct_alignment = None if ignore_alignment else struct_alignment
    lines.append("} " + _alignment_str(struct_alignment) + " ${prefix};")

    # Root level import creates an import loop.
    from reikna.cluda.kernel import Module

    return Module.create("\n".join(lines), render_kwds=kwds)


def ctype_module(dtype, ignore_alignment=False):
    """
    For a struct type, returns a :py:class:`~reikna.cluda.Module` object
    with the ``typedef`` of a struct corresponding to the given ``dtype``
    (with its name set to the module prefix);
    falls back to :py:func:`~reikna.cluda.dtypes.ctype` otherwise.

    The structure definition includes the alignment required
    to produce field offsets specified in ``dtype``;
    therefore, ``dtype`` must be either a simple type, or have
    proper offsets and dtypes (the ones that can be reporoduced in C
    using explicit alignment attributes, but without additional padding)
    and the attribute ``isalignedstruct == True``.
    An aligned dtype can be produced either by standard means
    (``aligned`` flag in ``numpy.dtype`` constructor and explicit offsets and itemsizes),
    or created out of an arbitrary dtype with the help of :py:func:`~reikna.cluda.dtypes.align`.

    If ``ignore_alignment`` is True, all of the above is ignored.
    The C structures produced will not have any explicit alignment modifiers.
    As a result, the the field offsets of ``dtype`` may differ from the ones
    chosen by the compiler.

    Modules are cached and the function returns a single module instance for equal ``dtype``'s.
    Therefore inside a kernel it will be rendered with the same prefix everywhere it is used.
    This results in a behavior characteristic for a structural type system,
    same as for the basic dtype-ctype conversion.

    .. warning::

        As of ``numpy`` 1.8, the ``isalignedstruct`` attribute is not enough to ensure
        a mapping between a dtype and a C struct with only the fields that are present in the dtype.
        Therefore, ``ctype_module`` will make some additional checks and raise ``ValueError``
        if it is not the case.
    """
    dtype = normalize_type(dtype)

    if dtype.names is None:
        return ctype(dtype)
    else:
        # FIXME: if numpy's ``isalignedstruct`` actually meant that the struct is aligned,
        # that would be enough.
        # Unfortunately, it only recognizes base alignments,
        # and does not check for itemsize consistency.
        # Therefore, there will be more checking in _get_struct_module.
        if not ignore_alignment and not dtype.isalignedstruct:
            raise ValueError("The data type must be an aligned struct")

        if len(dtype.shape) > 0:
            raise ValueError("The data type cannot be an array")

        if dtype not in _DTYPE_TO_CTYPE_MODULE:
            module = _get_struct_module(dtype, ignore_alignment=ignore_alignment)
            _DTYPE_TO_CTYPE_MODULE[dtype] = module
        else:
            module = _DTYPE_TO_CTYPE_MODULE[dtype]

        return module


def align(dtype):
    """
    Returns a new struct dtype with the field offsets changed to the ones a compiler would use
    (without being given any explicit alignment qualifiers).
    Ignores all existing explicit itemsizes and offsets.
    """
    dtype = normalize_type(dtype)

    if len(dtype.shape) > 0:
        return numpy.dtype((align(dtype.base), dtype.shape))

    if dtype.names is None:
        return dtype

    # Align the nested fields
    adjusted_fields = [
        align(dtype.fields[name][0])
        for name in dtype.names]

    # Get base alignments for the nested fields
    alignments = [_find_alignments(field_dtype)[0] for field_dtype in adjusted_fields]

    # Build offsets for the structure using a procedure
    # similar to the one a compiler would use
    offsets = [0]
    for name, prev_field_dtype, alignment in zip(
            dtype.names[1:], adjusted_fields[:-1], alignments[1:]):
        prev_end = offsets[-1] + prev_field_dtype.itemsize
        offsets.append(min_blocks(prev_end, alignment) * alignment)

    # Find the total itemsize.
    # According to the standard, it must be a multiple of the minimal alignment.
    struct_alignment = _struct_alignment(alignments)
    min_itemsize = offsets[-1] + adjusted_fields[-1].itemsize
    itemsize = min_blocks(min_itemsize, struct_alignment) * struct_alignment

    return numpy.dtype(dict(
        names=dtype.names,
        formats=adjusted_fields,
        offsets=offsets,
        itemsize=itemsize,
        aligned=True))


def _flatten_dtype(dtype, prefix=[]):

    if dtype.names is None:
        return [(prefix, dtype)]
    else:
        result = []
        for name in dtype.names:
            elem_dtype, _ = dtype.fields[name]
            if len(elem_dtype.shape) == 0:
                base_elem_dtype = elem_dtype
                elem_dtype_shape = tuple()
            else:
                base_elem_dtype = elem_dtype.base
                elem_dtype_shape = elem_dtype.shape

            if len(elem_dtype_shape) == 0:
                result += _flatten_dtype(base_elem_dtype, prefix=prefix + [name])
            else:
                for idxs in itertools.product(*[range(dim) for dim in elem_dtype_shape]):
                    result += _flatten_dtype(base_elem_dtype, prefix=prefix + [name] + list(idxs))
        return result


def flatten_dtype(dtype):
    """
    Returns a list of tuples ``(path, dtype)`` for each of the basic dtypes in
    a (possibly nested) ``dtype``.
    ``path`` is a list of field names/array indices leading to the corresponding element.
    """
    dtype = normalize_type(dtype)
    return _flatten_dtype(dtype)


def c_path(path):
    """
    Returns a string corresponding to the ``path`` to a struct element in C.
    The ``path`` is the sequence of field names/array indices returned from
    :py:func:`~reikna.cluda.dtypes.flatten_dtype`.
    """
    return  "".join(
        (("." + elem) if isinstance(elem, str) else ("[" + str(elem) + "]"))
        for elem in path)


def _extract_field(arr, path, array_idxs):
    """
    A helper function for ``extract_field``.
    Need to collect array indices for dtype sub-array fields since they are attached to the end
    of the full array index.
    """
    if len(path) == 0:
        if len(array_idxs) == 0:
            return arr
        else:
            slices = tuple(
                [slice(None, None, None)] * (len(arr.shape) - len(array_idxs)) + array_idxs)
            return arr[slices]
    elif isinstance(path[0], str):
        return _extract_field(arr[path[0]], path[1:], array_idxs)
    else:
        return _extract_field(arr, path[1:], array_idxs + [path[0]])


def extract_field(arr, path):
    """
    Extracts an element from an array of struct dtype.
    The ``path`` is the sequence of field names/array indices returned from
    :py:func:`~reikna.cluda.dtypes.flatten_dtype`.
    """
    return _extract_field(arr, path, [])



"""
This Py.Test plugin allows return value collection from testcases
(for cases when pass/fail is not enough).
"""

import pytest

# renderers
renderers = {
    'GFLOPS': lambda x: "{f:.2f} GFLOPS".format(f=float(x[1]) / x[0] / 1e9),
    'GB/s': lambda x: "{f:.2f} GB/s".format(f=float(x[1]) / x[0] / 2 ** 30)
}

def pytest_configure(config):
    config.pluginmanager.register(ReturnValuesPlugin(config), "returnvalues")
    config.addinivalue_line("markers", "returns(value_type): collect this testcase's return value")


class ReturnValuesPlugin(object):

    def __init__(self, config):
        pass

    def pytest_report_teststatus(self, __multicall__, report):
        outcome, letter, msg = __multicall__.execute()

        # if we have some result attached to the testcase, print it instead of 'PASSED'
        if hasattr(report, 'retval'):
            msg = report.retval

        return outcome, letter, msg

    def pytest_pyfunc_call(self, __multicall__, pyfuncitem):
        # collect testcase return result
        testfunction = pyfuncitem.obj

        # Taken from _pytest/python.py/pytest_pyfunc_call()
        # This bit uses some internal functions which seem to change without notice.
        # Need to replace it with proper mechanism when such functionality is available in pytest.
        if pyfuncitem._isyieldedfunction():
            res = testfunction(*pyfuncitem._args)
        else:
            funcargs = pyfuncitem.funcargs
            testargs = {}
            for arg in pyfuncitem._fixtureinfo.argnames:
                testargs[arg] = funcargs[arg]
            res = testfunction(**testargs)

        pyfuncitem.retval = res
        return True # finished processing the callback

    def pytest_runtest_makereport(self, __multicall__, item, call):
        report = __multicall__.execute()

        # if the testcase has passed, and has 'perf' marker, process its results
        if call.when == 'call' and report.passed and hasattr(item.function, 'returns'):
            mark = item.function.returns
            if len(mark.args) > 0:
                if mark.args[0] in renderers:
                    renderer = renderers[mark.args[0]]
                else:
                    renderer = lambda x: repr(x) + " " + mark.args[0]
            else:
                renderer = lambda x: repr(x)

            report.retval = renderer(item.retval)

        return report



"""
This example illustrates how to:
- define a custom structure type with a correct alignment
  (allowing it to be used to exchange data between numpy and kernels);
- construct a reduction computation that operates on arrays of the struct type above.
"""

import numpy
from reikna.cluda import dtypes, any_api
from reikna.algorithms import Reduce, Predicate
from reikna.cluda import Snippet
from reikna.core import Annotation, Type, Transformation, Parameter


# Pick the first available GPGPU API and make a Thread on it.
api = any_api()
thr = api.Thread.create()


# Minmax data type and the corresponding structure.
# Note that the dtype must be aligned before use on a GPU.
mmc_dtype = dtypes.align(numpy.dtype([
    ("cur_min", numpy.int32),
    ("cur_max", numpy.int32),
    ("pad", numpy.int32),
    ]))
mmc_c_decl = dtypes.ctype_module(mmc_dtype)


# Create the "empty" element for our minmax monoid, that is
# x `minmax` empty == empty `minmax` x == x.
empty = numpy.empty(1, mmc_dtype)[0]
empty["cur_min"] = 1 << 30
empty["cur_max"] = -(1 << 30)


# Reduction predicate for the minmax.
# v1 and v2 get the names of two variables to be processed.
predicate = Predicate(
    Snippet.create(lambda v1, v2: """
        ${ctype} result = ${v1};
        if (${v2}.cur_min < result.cur_min)
            result.cur_min = ${v2}.cur_min;
        if (${v2}.cur_max > result.cur_max)
            result.cur_max = ${v2}.cur_max;
        return result;
        """,
        render_kwds=dict(ctype=mmc_c_decl)),
    empty)


# Test array
arr = numpy.random.randint(0, 10**6, 20000)


# A transformation that creates initial minmax structures for the given array of integers
to_mmc = Transformation(
    [Parameter('output', Annotation(Type(mmc_dtype, arr.shape), 'o')),
    Parameter('input', Annotation(arr, 'i'))],
    """
    ${output.ctype} res;
    res.cur_min = ${input.load_same};
    res.cur_max = ${input.load_same};
    ${output.store_same}(res);
    """)


# Create the reduction computation and attach the transformation above to its input.
reduction = Reduce(to_mmc.output, predicate)
reduction.parameter.input.connect(to_mmc, to_mmc.output, new_input=to_mmc.input)
creduction = reduction.compile(thr)


# Run the computation
arr_dev = thr.to_device(arr)
res_dev = thr.empty_like(reduction.parameter.output)
creduction(res_dev, arr_dev)
minmax = res_dev.get()

assert minmax["cur_min"] == arr.min()
assert minmax["cur_max"] == arr.max()



from reikna.cluda import Snippet
import reikna.helpers as helpers
from reikna.core import Computation, Indices, Parameter, Annotation
from reikna.core.transformation import TransformationParameter


class PureParallel(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    A general class for pure parallel computations
    (i.e. with no interaction between threads).

    :param parameters: a list of :py:class:`~reikna.core.Parameter` objects.
    :param code: a source code for the computation.
        Can be a :py:class:`~reikna.cluda.Snippet` object which will be passed
        :py:class:`~reikna.core.Indices` object for the ``guiding_array`` as the first
        positional argument, and :py:class:`~reikna.core.transformation.KernelParameter` objects
        corresponding to ``parameters`` as the rest of positional arguments.
        If it is a string, such :py:class:`~reikna.cluda.Snippet` will be created out of it,
        with the parameter names ``idxs`` for the first one and the names of ``parameters``
        for the remaining ones.

    :param guiding_array: an tuple with the array shape, or the name of one of ``parameters``.
        By default, the first parameter is chosen.
    :param render_kwds: a dictionary with render keywords for the ``code``.

    .. py:function:: compiled_signature(*args)

        :param args: corresponds to the given ``parameters``.
    """

    def __init__(self, parameters, code, guiding_array=None, render_kwds=None):

        Computation.__init__(self, parameters)

        self._root_parameters = list(self.signature.parameters.keys())

        if isinstance(code, Snippet):
            self._snippet = code
        else:
            self._snippet = Snippet(helpers.template_def(
                ['idxs'] + self._root_parameters, code), render_kwds=render_kwds)

        if guiding_array is None:
            guiding_array = self._root_parameters[0]

        if isinstance(guiding_array, str):
            self._guiding_shape = self.signature.parameters[guiding_array].annotation.type.shape
        else:
            self._guiding_shape = guiding_array

    @classmethod
    def from_trf(cls, trf, guiding_array=None):
        """
        Creates a ``PureParallel`` instance from a :py:class:`~reikna.core.Transformation` object.
        ``guiding_array`` can be a string with a name of an array parameter from ``trf``,
        or the corresponding :py:class:`~reikna.core.transformation.TransformationParameter` object.
        """

        if guiding_array is None:
            guiding_array = trf.signature.parameters.keys()[0]

        if isinstance(guiding_array, TransformationParameter):
            if not guiding_array.belongs_to(trf):
                raise ValueError(
                    "The transformation parameter must belong to the provided transformation")
            guiding_array = str(guiding_array)

        guiding_param = trf.signature.parameters[guiding_array]
        if not guiding_param.annotation.array:
            raise ValueError("The parameter serving as a guiding array cannot be a scalar")

        # Transformation snippet is the same as required for PureParallel
        # In particular, it has arguments like "idxs, arg1, arg2, ...",
        # and load_same()/store_same() in the snippet will work because the PureParallel
        # computation defines index variables in its kernel (see _build_plan()).
        # This is a bit shady, but other variants are even worse
        # (e.g. creating a trivial computation and attaching this transformation to it
        # will either produce incorrect parameter order, or will require the usage of an
        # 'io' parameter, which has its own complications).
        # FIXME: find a solution which does not create an implicit dependence on
        # the way transformations are handled.
        res = cls(
            list(trf.signature.parameters.values()),
            trf.snippet,
            guiding_array=guiding_param.name)

        return res

    def _build_plan(self, plan_factory, _device_params, *args):

        plan = plan_factory()

        # Using root_parameters to avoid duplicated names
        # (can happen if this computation is nested and
        # the same arrays are passed to it as arguments)
        arglist = ", ".join(self._root_parameters)
        idxs = Indices(self._guiding_shape)

        template = helpers.template_def(
            ['kernel_declaration'] + self._root_parameters,
            """
            ${kernel_declaration}
            {
                VIRTUAL_SKIP_THREADS;

                %for i, idx in enumerate(idxs):
                VSIZE_T ${idx} = virtual_global_id(${i});
                %endfor

                ${snippet(idxs, """ + arglist + """)}
            }
            """)

        plan.kernel_call(
            template, args,
            global_size=(1,) if len(self._guiding_shape) == 0 else self._guiding_shape,
            render_kwds=dict(
                idxs=idxs,
                product=helpers.product,
                snippet=self._snippet))

        return plan



import sys
from collections import defaultdict
import itertools

from reikna.cluda import OutOfResourcesError
from reikna.cluda.kernel import render_template
from reikna.helpers import product, log2, min_blocks, template_for, wrap_in_tuple, factors

TEMPLATE = template_for(__file__)


# For effectiveness' sake, since we have quite a lot of range() usage here,
# and Py2's range() that returns list will be too slow.
_range = xrange if sys.version_info[0] < 3 else range


class PrimeFactors:
    """
    Contains a natural number's decomposition into prime factors.
    """

    def __init__(self, factors):
        self.factors = factors

    @classmethod
    def decompose(cls, num):
        factors = defaultdict(lambda: 0)
        if num == 1:
            return cls(dict(factors))

        while num > 1:
            for i in _range(2, int(round(num ** 0.5)) + 1):
                if num % i == 0:
                    factors[i] += 1
                    num //= i
                    break
            else:
                factors[num] += 1
                num = 1

        return cls(dict(factors))

    def get_value(self):
        res = 1
        for pwr, exp in self.factors.items():
            res *= pwr ** exp
        return res

    def get_arrays(self):
        return self.factors.keys(), self.factors.values()

    def div_by(self, other):
        factors = dict(self.factors)
        for o_pwr, o_exp in other.factors.items():
            factors[o_pwr] -= o_exp
            if factors[o_pwr] == 0:
                del factors[o_pwr]
        return PrimeFactors(factors)


def _get_decompositions(num_factors, parts):
    """
    Helper recursive function for ``get_decompositions()``.
    Iterates over all possible decompositions of ``num_factors`` (of type ``PrimeFactors``)
    into ``parts`` factors.
    """
    if parts == 1:
        yield (num_factors.get_value(),)
        return

    powers, exponents = num_factors.get_arrays()
    for sub_exps in itertools.product(*[_range(exp, -1, -1) for exp in exponents]):
        part_factors = PrimeFactors(
            dict(((pwr,sub_exp) for pwr, sub_exp in zip(powers, sub_exps) if sub_exp > 0)))
        part = part_factors.get_value()
        remainder = num_factors.div_by(part_factors)
        for decomp in _get_decompositions(remainder, parts - 1):
            yield (part,) + decomp


def get_decompositions(num, parts):
    """
    Iterates overall possible decompositions of ``num`` into ``parts`` factors.
    """
    num_factors = PrimeFactors.decompose(num)
    return _get_decompositions(num_factors, parts)


def find_local_size(global_size, flat_local_size, threshold=0.05):
    """
    Returns a tuple of the same size as ``global_size``,
    with the product equal to ``flat_local_size``,
    and minimal difference between ``product(global_size)``
    and ``product(min_blocks(gs, ls) for gs, ls in zip(global_size, local_size))``
    (i.e. tries to minimize the amount of empty threads).
    """
    flat_global_size = product(global_size)
    if flat_local_size >= flat_global_size:
        return global_size

    threads_num = product(global_size)

    best_ratio = None
    best_local_size = None

    for local_size in get_decompositions(flat_local_size, len(global_size)):
        bounding_global_size = tuple(
            ls * min_blocks(gs, ls) for gs, ls in zip(global_size, local_size))
        empty_threads = product(bounding_global_size) - threads_num
        ratio = float(empty_threads) / threads_num

        # Stopping iteration early, because there may be a lot of elements to iterate over,
        # and we do not need the perfect solution.
        if ratio < threshold:
            return local_size

        if best_ratio is None or ratio < best_ratio:
            best_ratio = ratio
            best_local_size = local_size

    return best_local_size


def _group_dimensions(vdim, virtual_shape, adim, available_shape):
    """
    ``vdim`` and ``adim`` are used for the absolute addressing of dimensions during recursive calls.
    """
    if len(virtual_shape) == 1 and virtual_shape[0] == 1:
        return [(vdim,)], [(adim,)]

    if len(virtual_shape) == 0:
        return [], []

    if virtual_shape[0] == 1:
        v_remainder, a_remainder = _group_dimensions(
            vdim + 1, virtual_shape[1:], adim, available_shape)
        return [(vdim,) + v_remainder[0]] + v_remainder[1:], a_remainder

    vdim_group = 1 # number of currently grouped virtual dimensions
    adim_group = 1 # number of currently grouped available dimensions

    while 1:
        # If we have more elements in the virtual group than there is in the available group,
        # extend the available group by one dimension.
        if product(virtual_shape[:vdim_group]) > product(available_shape[:adim_group]):
            adim_group += 1
            continue

        # If the remaining available dimensions cannot accommodate the remaining virtual dimensions,
        # we try to fit one more virtual dimension in the virtual group.
        if product(virtual_shape[vdim_group:]) > product(available_shape[adim_group:]):
            vdim_group += 1
            continue

        # If we are here, it means that:
        # 1) the current available group can accommodate the current virtual group;
        # 2) the remaining available dimensions can accommodate the remaining virtual dimensions.
        # This means we can make a recursive call now.
        v_res = tuple(range(vdim, vdim + vdim_group))
        a_res = tuple(range(adim, adim + adim_group))
        v_remainder, a_remainder = _group_dimensions(
            vdim + vdim_group, virtual_shape[vdim_group:],
            adim + adim_group, available_shape[adim_group:])
        return [v_res] + v_remainder, [a_res] + a_remainder


def group_dimensions(virtual_shape, available_shape):
    """
    Returns two lists, one of tuples with numbers of grouped virtual dimensions, the other
    one of tuples with numbers of corresponding group of available dimensions,
    such that for any group of virtual dimensions, the total number of elements they cover
    does not exceed the number of elements covered by the
    corresponding group of available dimensions.
    """
    assert product(virtual_shape) <= product(available_shape)
    return _group_dimensions(0, virtual_shape, 0, available_shape)


def ceiling_root(num, pwr):
    """
    Returns the integer ``num ** (1. / pwr)`` if num is a perfect square/cube/etc,
    or the integer ceiling of this value, if it's not.
    """
    res = num ** (1. / pwr)
    int_res = int(round(res))
    if int_res ** pwr == num:
        return int_res
    else:
        return int(res + 1)


def find_bounding_shape(virtual_size, available_shape):
    """
    Finds a tuple of the same length as ``available_shape``, with every element
    not greater than the corresponding element of ``available_shape``,
    and product not lower than ``virtual_size``.
    """
    assert virtual_size <= product(available_shape)

    free_size = virtual_size
    free_dims = set(range(len(available_shape)))
    bounding_shape = [None] * len(available_shape)

    while len(free_dims) > 0:
        guess = ceiling_root(free_size, len(free_dims))
        for fdim in free_dims:
            bounding_shape[fdim] = guess

        for fdim in free_dims:
            if bounding_shape[fdim] > available_shape[fdim]:
                bounding_shape[fdim] = available_shape[fdim]
                free_dims.remove(fdim)
                free_size = min_blocks(free_size, bounding_shape[fdim])
                break
        else:
            return tuple(bounding_shape)

    return tuple(available_shape)


class ShapeGroups:

    def __init__(self, virtual_shape, available_shape):
        self.real_dims = {}
        self.real_strides = {}
        self.virtual_strides = {}
        self.major_vdims = {}
        self.bounding_shape = tuple()
        self.skip_thresholds = []

        v_groups, a_groups = group_dimensions(virtual_shape, available_shape)

        for v_group, a_group in zip(v_groups, a_groups):
            virtual_subshape = virtual_shape[v_group[0]:v_group[-1]+1]
            virtual_subsize = product(virtual_subshape)

            bounding_subshape = find_bounding_shape(
                virtual_subsize,
                available_shape[a_group[0]:a_group[-1]+1])

            self.bounding_shape += bounding_subshape

            if virtual_subsize < product(bounding_subshape):
                strides = [(adim, product(bounding_subshape[:i])) for i, adim in enumerate(a_group)]
                self.skip_thresholds.append((virtual_subsize, strides))

            for vdim in v_group:
                self.real_dims[vdim] = a_group
                self.real_strides[vdim] = tuple(
                    product(self.bounding_shape[a_group[0]:adim]) for adim in a_group)
                self.virtual_strides[vdim] = product(virtual_shape[v_group[0]:vdim])
                self.major_vdims[vdim] = v_group[-1]


class VirtualSizes:

    def __init__(self, device_params, virtual_global_size,
            virtual_local_size=None, max_local_size=None):

        virtual_global_size = wrap_in_tuple(virtual_global_size)
        if virtual_local_size is not None:
            virtual_local_size = wrap_in_tuple(virtual_local_size)
            if len(virtual_local_size) != len(virtual_global_size):
                raise ValueError("Global size and local size must have the same dimensions")

        # Since the device uses column-major ordering of sizes, while we get
        # row-major ordered shapes, we temporarily invert our shapes
        # to facilitate internal handling.
        virtual_global_size = tuple(reversed(virtual_global_size))
        if virtual_local_size is not None:
            virtual_local_size = tuple(reversed(virtual_local_size))

        # Restrict local sizes using the provided explicit limit
        if max_local_size is not None:
            max_work_group_size = min(
                max_local_size,
                device_params.max_work_group_size,
                product(device_params.max_work_item_sizes))
            max_work_item_sizes = [
                min(max_local_size, mwis) for mwis in device_params.max_work_item_sizes]
        else:
            # Assuming:
            # 1) max_work_group_size <= product(max_work_item_sizes)
            # 2) max(max_work_item_sizes) <= max_work_group_size
            max_work_group_size = device_params.max_work_group_size
            max_work_item_sizes = device_params.max_work_item_sizes

        if virtual_local_size is None:
            # FIXME: we can obtain better results by taking occupancy into account here,
            # but for now we will assume that the more threads, the better.
            flat_global_size = product(virtual_global_size)
            multiple = device_params.warp_size

            if flat_global_size < max_work_group_size:
                flat_local_size = flat_global_size
            elif max_work_group_size < multiple:
                flat_local_size = 1
            else:
                flat_local_size = multiple * (max_work_group_size // multiple)

            # product(virtual_local_size) == flat_local_size <= max_work_group_size
            virtual_local_size = find_local_size(virtual_global_size, flat_local_size)
        else:
            if product(virtual_local_size) > max_work_group_size:
                raise OutOfResourcesError(
                    "Requested local size is greater than the maximum " + str(max_work_group_size))

        # Global and local sizes supported by CUDA or OpenCL restricted number of dimensions,
        # which may have limited size, so we need to pack our multidimensional sizes.

        virtual_grid_size = tuple(
            min_blocks(gs, ls) for gs, ls in zip(virtual_global_size, virtual_local_size))
        bounding_global_size = tuple(
            grs * ls for grs, ls in zip(virtual_grid_size, virtual_local_size))

        if product(virtual_grid_size) > product(device_params.max_num_groups):
            raise OutOfResourcesError(
                "Bounding global size " + repr(bounding_global_size) + " is too large")

        local_groups = ShapeGroups(virtual_local_size, max_work_item_sizes)
        grid_groups = ShapeGroups(virtual_grid_size, device_params.max_num_groups)

        # Returning back to the row-major ordering
        self.virtual_local_size = tuple(reversed(virtual_local_size))
        self.virtual_global_size = tuple(reversed(virtual_global_size))

        # These can be different lenghts because of expansion into multiple dimensions
        # find_bounding_shape() does.
        real_local_size = tuple(local_groups.bounding_shape)
        real_grid_size = tuple(grid_groups.bounding_shape)

        diff = len(real_local_size) - len(real_grid_size)
        if diff > 0:
            self.real_local_size = real_local_size
            self.real_grid_size = real_grid_size + (1,) * abs(diff)
        else:
            self.real_local_size = real_local_size + (1,) * abs(diff)
            self.real_grid_size = real_grid_size

        self.real_global_size = tuple(
            gs * ls for gs, ls
            in zip(self.real_grid_size, self.real_local_size))

        # This function will be used to translate between internal column-major vdims
        # and user-supplied row-major vdims.
        vdim_inverse = lambda dim: len(self.virtual_local_size) - dim - 1

        self.vsize_functions = render_template(
            TEMPLATE,
            virtual_local_size=virtual_local_size,
            virtual_global_size=virtual_global_size,
            bounding_global_size=bounding_global_size,
            virtual_grid_size=virtual_grid_size,
            local_groups=local_groups,
            grid_groups=grid_groups,
            product=product,
            vdim_inverse=vdim_inverse)



import numpy
from numpy.polynomial import Hermite

import reikna.cluda.dtypes as dtypes
from reikna.core import Computation, Parameter, Annotation, Type
from reikna.algorithms import Transpose
from reikna.linalg import MatrixMul


def factorial(num):
    """
    Replacement of ``scipy.misc.factorial()``.
    First, it avoids requiring scipy; second, it returns integer instead of float
    (which gives more precision for subsequent calculations).
    """
    res = 1
    for i in range(2, num + 1):
        res *= i
    return res


def hermite(mode):
    """Returns an orthonormal Hermite polynomial"""
    def func(x_coord):
        norm = 1. / (numpy.pi ** 0.25) / numpy.sqrt(float(factorial(mode) * 2 ** mode))
        return Hermite([0] * mode + [1])(x_coord) * norm

    return func


def h_roots(order):
    """
    Recursive root finding algorithm, taken from Numerical Recipes.
    More accurate than the standard h_roots() from scipy.
    """

    eps = 1.0e-14
    pim4 = numpy.pi ** (-0.25)
    max_iter = 20 # Maximum iterations.

    roots = numpy.empty(order)
    weights = numpy.empty(order)

    # Initial guess for the largest root
    curr_root = numpy.sqrt(2 * order + 1) - 1.85575 * (2 * order + 1) ** (-0.16667)

    for i in range((order + 1) // 2):

        # Initial guesses for the following roots
        if i == 1:
            curr_root -= 1.14 * order ** 0.426 / curr_root
        elif i == 2:
            curr_root = 1.86 * curr_root + 0.86 * roots[0]
        elif i == 3:
            curr_root = 1.91 * curr_root + 0.91 * roots[1]
        elif i > 3:
            curr_root = 2.0 * curr_root + roots[i - 2]

        # Refinement by Newton's method
        for _ in range(max_iter):
            pval = pim4
            pval_prev = 0.0

            # Recurrence relation to get the Hermite polynomial evaluated at ``curr_root``
            for j in range(order):
                pval, pval_prev = (curr_root * numpy.sqrt(2.0 / (j + 1)) * pval -
                    numpy.sqrt(float(j) / (j + 1)) * pval_prev), pval

            # ``pval`` is now the desired Hermite polynomial
            # We next compute ``pderiv``, its derivative, using ``pval_prev``, the polynomial
            # of one lower order.
            pderiv = numpy.sqrt(2 * order) * pval_prev
            prev_root = curr_root
            curr_root -= pval / pderiv
            if abs(curr_root - prev_root) <= eps:
                break
        else:
            raise Exception("Too many iterations")

        roots[order - 1 - i] = curr_root
        roots[i] = -curr_root
        weights[i] = 2.0 / (pderiv ** 2)
        weights[order - 1 - i] = weights[i]

    return roots, weights


def get_spatial_points(modes, order, add_points=0):
    """
    Returns the number of points in coordinate space
    which allows the calculation of up to ``order`` of any function
    initially defined on mode space.
    ``add_points`` can be used to increase the size of the grid.
    """
    points = ((modes - 1) * (order + 1)) // 2 + 1

    # TODO: Population calculated in mode and in x-space is slightly different
    # The more points we take in addition to minimum necessary for precise
    # G.-H. quadrature, the less is the difference.
    # Looks like it is not a bug, just inability to integrate Hermite function
    # in x-space precisely (oscillates too fast maybe?).
    # But this still requres investigation.
    # In addition: with dp != 0 X-M-X transform of TF state
    # gives non-smooth curve. Certainly TF state has some higher harmonics
    # (infinite number of them, to be precise), but why it is smooth when dp = 0?
    points += add_points

    return points


def get_spatial_grid_and_weights(modes, order, add_points=0):
    """
    Returns a pair of arrays ``(points, weights)`` for Gauss-Hermite quadrature.
    """
    points = get_spatial_points(modes, order, add_points=add_points)
    roots, weights = h_roots(points)

    return roots * numpy.sqrt(2.0 / (order + 1)), \
        weights * numpy.exp(roots ** 2) * numpy.sqrt(2.0 / (order + 1))


def get_spatial_grid(modes, order, add_points=0):
    """
    Returns the spatial grid required to calculate the ``order`` power of a function
    defined in the harmonic mode space of the size ``modes``.
    If ``add_points`` is 0, the grid has the minimum size required for exact
    transformation back to the mode space.
    """
    return get_spatial_grid_and_weights(modes, order, add_points=add_points)[0]


def get_spatial_weights(modes, order, add_points=0):
    return get_spatial_grid_and_weights(modes, order, add_points=add_points)[1]


def harmonic(mode):
    r"""
    Returns an eigenfunction of order :math:`n = \mathrm{mode}` for the harmonic oscillator:

    .. math::
        \phi_{n}
        = \frac{1}{\sqrt[4]{\pi} \sqrt{2^n n!}} H_n(x) \exp(-x^2/2),

    where :math:`H_n` is the :math:`n`-th order "physicists'" Hermite polynomial.
    The normalization is chosen so that :math:`\int \phi_n^2(x) dx = 1`.
    """
    polynomial = hermite(mode)
    return lambda x_coord: polynomial(x_coord) * numpy.exp(-(x_coord ** 2) / 2)


def get_transformation_matrix(modes, order, add_points):
    """
    Returns the the matrix of values of mode functions taken at
    points of the spatial grid.
    """
    x_coords = get_spatial_grid(modes, order, add_points=add_points)

    res = numpy.zeros((modes, x_coords.size))

    for mode in range(modes):
        res[mode, :] = harmonic(mode)(x_coords)

    return res


class DHT(Computation):
    r"""
    Bases: :py:class:`~reikna.core.Computation`

    Discrete transform to and from harmonic oscillator modes.
    With ``inverse=True`` transforms a function defined by its expansion
    :math:`C_m,\,m=0 \ldots M-1` in the mode space with mode functions
    from :py:func:`~reikna.dht.harmonic`,
    to the coordinate space (:math:`F(x)` on the grid :math:`x`
    from :py:func:`~reikna.dht.get_spatial_grid`).
    With ``inverse=False`` guarantees to recover first :math:`M` modes of :math:`F^k(x)`,
    where :math:`k` is the ``order`` parameter.

    For multiple dimensions the operation is the same,
    and the mode functions are products of 1D mode functions, i.e.
    :math:`\phi_{l,m,n}^{3D}(x,y,z) = \phi_l(x) \phi_m(y) \phi_n(z)`.

    For the detailed description of the algorithm, see Dion & Cances,
    `PRE 67(4) 046706 (2003) <http://dx.doi.org/10.1103/PhysRevE.67.046706>`_

    :param mode_arr: an array-like object defining the shape of mode space.
        If ``inverse=False``, its shape is used to define the mode space size.
    :param inverse: ``False`` for forward (coordinate space -> mode space) transform,
        ``True`` for inverse (mode space -> coordinate space) transform.
    :param axes: a tuple with axes over which to perform the transform.
        If not given, the transform is performed over all the axes.
    :param order: if ``F`` is a function in mode space, the number of spatial points
        is chosen so that the transformation ``DHT[(DHT^{-1}[F])^order]`` could be performed.
    :param add_points: a list of the same length as ``mode_arr`` shape,
        specifying the number of points in x-space to use in addition to minimally required
        (``0`` by default).

    .. py:method:: compiled_signature_forward(modes:o, coords:i)
    .. py:method:: compiled_signature_inverse(coords:o, modes:i)

        Depending on ``inverse`` value, either of these two will be created.

        :param modes: an array with the attributes of ``mode_arr``.
        :param coords: an array with the shape depending on ``mode_arr``, ``axes``, ``order``
            and ``add_points``, and the dtype of ``mode_arr``.
    """

    def __init__(self, mode_arr, add_points=None, inverse=False, order=1, axes=None):

        if axes is None:
            axes = tuple(range(len(mode_arr.shape)))
        else:
            axes = tuple(axes)
        self._axes = list(sorted(axes))

        if add_points is None:
            add_points = [0] * len(mode_arr.shape)
        else:
            add_points = list(add_points)
        self._add_points = add_points

        coord_shape = list(mode_arr.shape)
        for axis in range(len(mode_arr.shape)):
            if axis in axes:
                coord_shape[axis] = get_spatial_points(
                    mode_arr.shape[axis], order, add_points=add_points[axis])
        coord_arr = Type(mode_arr.dtype, shape=coord_shape)

        self._inverse = inverse
        self._order = order

        if not inverse:
            parameters = [
                Parameter('modes', Annotation(mode_arr, 'o')),
                Parameter('coords', Annotation(coord_arr, 'i'))]
        else:
            parameters = [
                Parameter('coords', Annotation(coord_arr, 'o')),
                Parameter('modes', Annotation(mode_arr, 'i'))]

        Computation.__init__(self, parameters)

    def _get_transformation_matrix(self, dtype, modes, add_points):
        p_matrix = get_transformation_matrix(modes, self._order, add_points)
        p_matrix = p_matrix.astype(dtype)

        if not self._inverse:
            weights = get_spatial_weights(modes, self._order, add_points)
            tiled_weights = numpy.tile(
                weights.reshape(weights.size, 1).astype(dtype),
                (1, modes))
            p_matrix = p_matrix.transpose() * tiled_weights

        return p_matrix

    def _add_transpose(self, plan, current_mem, current_axes, axis):
        """
        Transpose the current array so that the ``axis`` is in the end of axes list.
        """

        seq_axes = list(range(len(current_axes)))

        cur_pos = current_axes.index(axis)
        if cur_pos != len(current_axes) - 1:

            # We can move the target axis to the end in different ways,
            # but this one will require only one transpose kernel.
            optimal_transpose = lambda seq: seq[:cur_pos] + seq[cur_pos+1:] + [seq[cur_pos]]

            tr_axes = optimal_transpose(seq_axes)
            new_axes = optimal_transpose(current_axes)

            transpose = Transpose(current_mem, axes=tr_axes)
            tr_output = plan.temp_array_like(transpose.parameter.output)
            plan.computation_call(transpose, tr_output, current_mem)

            current_mem = tr_output
            current_axes = new_axes

        return current_mem, current_axes

    def _build_plan(self, plan_factory, _device_params, output_arr, input_arr):

        plan = plan_factory()

        dtype = input_arr.dtype
        p_dtype = dtypes.real_for(dtype) if dtypes.is_complex(dtype) else dtype

        mode_shape = input_arr.shape if self._inverse else output_arr.shape

        current_mem = input_arr
        seq_axes = list(range(len(input_arr.shape)))
        current_axes = list(range(len(input_arr.shape)))

        for i, axis in enumerate(self._axes):
            current_mem, current_axes = self._add_transpose(plan, current_mem, current_axes, axis)

            tr_matrix = plan.persistent_array(
                self._get_transformation_matrix(p_dtype, mode_shape[axis], self._add_points[axis]))

            dot = MatrixMul(current_mem, tr_matrix)
            if i == len(self._axes) - 1 and current_axes == seq_axes:
                dot_output = output_arr
            else:
                # Cannot write to output if it is not the last transform,
                # or if we need to return to the initial axes order
                dot_output = plan.temp_array_like(dot.parameter.output)
            plan.computation_call(dot, dot_output, current_mem, tr_matrix)
            current_mem = dot_output

        # If we ended up with the wrong order of axes,
        # return to the original order.

        if current_axes != seq_axes:
            tr_axes = [current_axes.index(i) for i in range(len(current_axes))]
            transpose = Transpose(current_mem, axes=tr_axes)
            plan.add_computation(transpose, output_arr, current_mem)

        return plan



import numpy
import pytest

from helpers import *
from reikna.algorithms import PureParallel
from reikna.core import Parameter, Annotation, Type, Computation
import reikna.cluda.dtypes as dtypes
from reikna.transformations import mul_param, copy


class NestedPureParallel(Computation):

    def __init__(self, size, dtype):

        Computation.__init__(self, [
            Parameter('output', Annotation(Type(dtype, shape=size), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=size), 'i'))])

        self._p = PureParallel([
                Parameter('output', Annotation(Type(dtype, shape=size), 'o')),
                Parameter('i1', Annotation(Type(dtype, shape=size), 'i')),
                Parameter('i2', Annotation(Type(dtype, shape=size), 'i'))],
            """
            ${i1.ctype} t1 = ${i1.load_idx}(${idxs[0]});
            ${i2.ctype} t2 = ${i2.load_idx}(${idxs[0]});
            ${output.store_idx}(${idxs[0]}, t1 + t2);
            """)

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        plan.computation_call(self._p, output, input_, input_)
        return plan


def test_nested(thr):

    N = 1000
    dtype = numpy.float32

    p = NestedPureParallel(N, dtype)

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = a + a

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_input(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=(2, N)), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=N), 'i'))],
        """
        float t = ${input.load_idx}(${idxs[0]});
        ${output.store_idx}(0, ${idxs[0]}, t);
        ${output.store_idx}(1, ${idxs[0]}, t * 2);
        """,
        guiding_array='input')

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = numpy.vstack([a, a * 2])

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_output(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=N), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=(2, N)), 'i'))],
        """
        float t1 = ${input.load_idx}(0, ${idxs[0]});
        float t2 = ${input.load_idx}(1, ${idxs[0]});
        ${output.store_idx}(${idxs[0]}, t1 + t2);
        """,
        guiding_array='output')

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = a[0] + a[1]

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_shape(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=(2, N)), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=(2, N)), 'i'))],
        """
        float t1 = ${input.load_idx}(0, ${idxs[0]});
        float t2 = ${input.load_idx}(1, ${idxs[0]});
        ${output.store_idx}(0, ${idxs[0]}, t1 + t2);
        ${output.store_idx}(1, ${idxs[0]}, t1 - t2);
        """,
        guiding_array=(N,))

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = numpy.vstack([a[0] + a[1], a[0] - a[1]])

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_zero_length_shape(thr):

    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=tuple()), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=tuple()), 'i'))],
        """
        float t = ${input.load_idx}();
        ${output.store_idx}(t * 2);
        """,
        guiding_array=tuple())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = (a * 2).astype(dtype)

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_trf_with_guiding_input(thr):
    """
    Test the creation of ``PureParallel`` out of a transformation,
    with an input parameter as a guiding array.
    """

    N = 1000
    coeff = 3
    dtype = numpy.float32

    arr_t = Type(dtype, shape=N)
    trf = mul_param(arr_t, dtype)
    p = PureParallel.from_trf(trf, trf.input)

    # The new PureParallel has to preserve the parameter list of the original transformation.
    assert list(p.signature.parameters.values()) == list(trf.signature.parameters.values())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev, coeff)

    assert diff_is_negligible(res_dev.get(), a * 3)


def test_trf_with_guiding_output(thr):
    """
    Test the creation of ``PureParallel`` out of a transformation,
    with an output parameter as a guiding array.
    """

    N = 1000
    coeff = 3
    dtype = numpy.float32

    arr_t = Type(dtype, shape=N)
    trf = mul_param(arr_t, dtype)
    p = PureParallel.from_trf(trf, trf.output)

    # The new PureParallel has to preserve the parameter list of the original transformation.
    assert list(p.signature.parameters.values()) == list(trf.signature.parameters.values())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev, coeff)

    assert diff_is_negligible(res_dev.get(), a * 3)


class TestSameArgument(Computation):

    def __init__(self, arr):

        copy_trf = copy(arr, out_arr_t=arr)
        self._copy_comp = PureParallel.from_trf(copy_trf, copy_trf.input)

        Computation.__init__(self, [
            Parameter('outer_output', Annotation(arr, 'o')),
            Parameter('outer_input', Annotation(arr, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        temp = plan.temp_array_like(output)
        plan.computation_call(self._copy_comp, temp, input_)
        plan.computation_call(self._copy_comp, temp, temp)
        plan.computation_call(self._copy_comp, output, temp)
        return plan


def test_same_argument(some_thr):
    """
    A regression test for an unexpected interaction of the way PureParallel.from_trf() worked
    and a logic flaw in processing 'io'-type nodes in a transformation tree.

    from_trf() created a trivial computation with a single 'io' parameter and
    attached the given transformation to it.
    This preserved the order of parameters in the resultinc computation
    (because of how the transformation tree was traversed),
    and quite nicely relied only on the public API.

    So, for a simple transformation with one input and one output the root PP computation
    looked like:

        input (io)

    and after attaching the transformation:

        input (i)
        input (o) ---(tr:input -> tr:output)---> output

    When this computation was used inside another computation and was passed the same argument
    (e.g. 'temp') both for input and output, during the translation stage
    this would be transformed to (since 'temp' is passed both to 'input' and 'output')

        temp (i)
        temp (o) ---(tr:input -> tr:output)---> temp

    because the translation was purely name-based.
    This resulted in some cryptic errors due to the name clash.
    Now the masked 'input' should have been mangled instead of translated,
    producing something like

        temp (i)
        _nested_input (o) ---(tr:input -> tr:output)---> temp

    but this functionality was not implemented.
    """
    arr = get_test_array((1000, 8, 1), numpy.complex64)
    arr_dev = some_thr.to_device(arr)

    test = TestSameArgument(arr)
    testc = test.compile(some_thr)

    testc(arr_dev, arr_dev)

    assert diff_is_negligible(arr_dev.get(), arr)



import numpy

import reikna.helpers as helpers
from reikna.core import Computation, Parameter, Annotation
from reikna.cluda import functions
import reikna.cluda.dtypes as dtypes
from reikna.cluda import OutOfResourcesError
from reikna.algorithms import PureParallel
from reikna.transformations import copy

TEMPLATE = helpers.template_for(__file__)


MAX_RADIX = 16


def get_radix_array(size, use_max_radix=False):
    """
    For any ``size``, this function decomposes ``size`` into factors for loacal memory tranpose
    based fft. Factors (radices) are sorted such that the first one (radix_array[0])
    is the largest. This base radix determines the number of registers used by each
    work item and product of remaining radices determine the size of work group needed.
    To make things concrete with and example, suppose ``size`` = 1024. It is decomposed into
    1024 = 16 x 16 x 4. Hence kernel uses float2 a[16], for local in-register fft and
    needs 16 x 4 = 64 work items per work group. So kernel first performance 64 length
    16 ffts (64 work items working in parallel) following by transpose using local
    memory followed by again 64 length 16 ffts followed by transpose using local memory
    followed by 256 length 4 ffts. For the last step since with size of work group is
    64 and each work item can array for 16 values, 64 work items can compute 256 length
    4 ffts by each work item computing 4 length 4 ffts.
    Similarly for ``size`` = 2048 = 8 x 8 x 8 x 4, each work group has 8 x 8 x 4 = 256 work
    iterms which each computes 256 (in-parallel) length 8 ffts in-register, followed
    by transpose using local memory, followed by 256 length 8 in-register ffts, followed
    by transpose using local memory, followed by 256 length 8 in-register ffts, followed
    by transpose using local memory, followed by 512 length 4 in-register ffts. Again,
    for the last step, each work item computes two length 4 in-register ffts and thus
    256 work items are needed to compute all 512 ffts.
    For ``size`` = 32 = 8 x 4, 4 work items first compute 4 in-register
    lenth 8 ffts, followed by transpose using local memory followed by 8 in-register
    length 4 ffts, where each work item computes two length 4 ffts thus 4 work items
    can compute 8 length 4 ffts. However if work group size of say 64 is choosen,
    each work group can compute 64/ 4 = 16 size 32 ffts (batched transform).
    Users can play with these parameters to figure what gives best performance on
    their particular device i.e. some device have less register space thus using
    smaller base radix can avoid spilling ... some has small local memory thus
    using smaller work group size may be required etc
    """
    assert size == 2 ** helpers.log2(size)

    if use_max_radix:
        radix = min(size, MAX_RADIX)
        radix_array = []
        while size > radix:
            radix_array.append(radix)
            size //= radix
        radix_array.append(size)
        return radix_array
    else:
        arrays = {
            2: [2], 4: [4], 8: [8],
            16: [8, 2], 32: [8, 4], 64: [8, 8],
            128: [8, 4, 4],
            256: [4, 4, 4, 4],
            512: [8, 8, 8],
            1024: [16, 16, 4],
            2048: [8, 8, 8, 4]
        }
        if size in arrays:
            return arrays[size]
        else:
            # Naive algorithm, can be imroved.
            lsize = helpers.log2(size)
            num_elems = helpers.min_blocks(lsize, 4)
            return [16] * (num_elems - 1) + [16 if lsize % 4 == 0 else 2 ** (lsize % 4)]


def get_global_radix_info(size):
    """
    For ``size`` larger than what can be computed using local memory fft, global transposes
    multiple kernel launces is needed. For these sizes, ``size`` can be decomposed using
    much larger base radices i.e. say ``size`` = 262144 = 128 x 64 x 32. Thus three kernel
    launches will be needed, first computing 64 x 32, length 128 ffts, second computing
    128 x 32 length 64 ffts, and finally a kernel computing 128 x 64 length 32 ffts.
    Each of these base radices can futher be divided into factors so that each of these
    base ffts can be computed within one kernel launch using in-register ffts and local
    memory transposes i.e for the first kernel above which computes 64 x 32 ffts on length
    128, 128 can be decomposed into 128 = 16 x 8 i.e. 8 work items can compute 8 length
    16 ffts followed by transpose using local memory followed by each of these eight
    work items computing 2 length 8 ffts thus computing 16 length 8 ffts in total. This
    means only 8 work items are needed for computing one length 128 fft. If we choose
    work group size of say 64, we can compute 64/8 = 8 length 128 ffts within one
    work group. Since we need to compute 64 x 32 length 128 ffts in first kernel, this
    means we need to launch 64 x 32 / 8 = 256 work groups with 64 work items in each
    work group where each work group is computing 8 length 128 ffts where each length
    128 fft is computed by 8 work items. Same logic can be applied to other two kernels
    in this example. Users can play with difference base radices and difference
    decompositions of base radices to generates different kernels and see which gives
    best performance. Following function is just fixed to use 128 as base radix
    """
    assert size == 2 ** helpers.log2(size)

    base_radix = min(size, 128)

    num_radices = 0
    while size > base_radix:
        size //= base_radix
        num_radices += 1

    radix_list = [base_radix] * num_radices + [size]
    radix1_list = []
    radix2_list = []

    for radix in radix_list:
        if radix <= 8:
            radix1_list.append(radix)
            radix2_list.append(1)
        else:
            radix1 = 2
            radix2 = radix // radix1
            while radix2 > radix1:
                radix1 *= 2
                radix2 = radix // radix1

            radix1_list.append(radix1)
            radix2_list.append(radix2)

    # sanity checks:
    for radix, radix1, radix2 in zip(radix_list, radix1_list, radix2_list):
        assert radix2 <= radix1
        assert radix1 * radix2 == radix
        assert radix1 <= MAX_RADIX

    return radix_list, radix1_list, radix2_list


def get_padding(threads_per_xform, radix_prev, threads_req, xforms_per_workgroup, radix, num_banks):

    if threads_per_xform <= radix_prev or radix_prev >= num_banks:
        offset = 0
    else:
        num_rows_req = ((threads_per_xform if threads_per_xform < num_banks else num_banks) //
            radix_prev)
        num_cols_req = 1
        if num_rows_req > radix:
            num_cols_req = num_rows_req // radix
        num_cols_req *= radix_prev
        offset = num_cols_req

    if threads_per_xform >= num_banks or xforms_per_workgroup == 1:
        mid_pad = 0
    else:
        bank_num = ((threads_req + offset) * radix) & (num_banks - 1)
        if bank_num >= threads_per_xform:
            mid_pad = 0
        else:
            # TODO: find out which conditions are necessary to execute this code
            mid_pad = threads_per_xform - bank_num

    lmem_size = ((threads_req + offset) * radix * xforms_per_workgroup +
        mid_pad * (xforms_per_workgroup - 1))
    return lmem_size, offset, mid_pad


def get_local_memory_size(size, radix_array, threads_per_xform, xforms_per_workgroup,
        num_local_mem_banks, min_mem_coalesce_width):

    lmem_size = 0

    # from insertGlobal(Loads/Stores)AndTranspose
    if threads_per_xform < min_mem_coalesce_width:
        lmem_size = max(lmem_size, (size + threads_per_xform) * xforms_per_workgroup)

    radix_prev = 1
    len_ = size
    for i, radix in enumerate(radix_array):

        threads_req = size // radix
        radix_curr = radix_prev * radix

        if i < len(radix_array) - 1:
            lmem_size_new, _, _ = get_padding(
                threads_per_xform, radix_prev, threads_req, xforms_per_workgroup,
                radix, num_local_mem_banks)
            lmem_size = max(lmem_size, lmem_size_new)
            radix_prev = radix_curr
            len_ = len_ // radix

    return lmem_size


def get_kweights(size_real, size_bound):
    """
    Returns weights to be applied as a part of Bluestein's algorithm
    between forward and inverse FFTs.
    """

    args = lambda ns: 1j * numpy.pi / size_real * ns ** 2

    n_v = numpy.concatenate([
        numpy.arange(size_bound - size_real +1),
        numpy.arange(size_real - 1, 0, -1)])

    return numpy.vstack([
        numpy.fft.fft(numpy.exp(args(n_v))),
        numpy.fft.ifft(numpy.exp(-args(n_v))) * size_bound / size_real])


def get_common_kwds(dtype, device_params):
    return dict(
        dtype=dtype,
        min_mem_coalesce_width=device_params.min_mem_coalesce_width[dtype.itemsize],
        local_mem_banks=device_params.local_mem_banks,
        get_padding=get_padding,
        wrap_const=lambda x: dtypes.c_constant(x, dtypes.real_for(dtype)),
        min_blocks=helpers.min_blocks,
        mul=functions.mul(dtype, dtype),
        polar_unit=functions.polar_unit(dtypes.real_for(dtype)),
        cdivs=functions.div(dtype, numpy.uint32, out_dtype=dtype))


class LocalFFTKernel:
    """Generator for 'local' FFT in shared memory"""

    def __init__(self, dtype, device_params, outer_shape, fft_size, fft_size_real,
            inner_shape, reverse_direction):

        self.name = "fft_local"
        self.inplace_possible = True
        self.output_shape = outer_shape + (fft_size_real if reverse_direction else fft_size,)
        if fft_size_real != fft_size and reverse_direction:
            self.kweights = get_kweights(fft_size_real, fft_size)
        else:
            self.kweights = None

        self._fft_size = fft_size
        self._fft_size_real = fft_size_real
        self._outer_batch = helpers.product(outer_shape)
        self._local_mem_size = device_params.local_mem_size
        self._itemsize = dtype.itemsize

        self._constant_kwds = get_common_kwds(dtype, device_params)
        self._constant_kwds.update(dict(
            takes_kweights=(self.kweights is not None),
            input_slices=(len(outer_shape), 1, len(inner_shape)),
            output_slices=(len(outer_shape), 1, len(inner_shape)),
            pad_in=(fft_size != fft_size_real and not reverse_direction),
            unpad_out=(fft_size != fft_size_real and reverse_direction),
            reverse_direction=reverse_direction,
            normalize=True))

    def prepare_for(self, max_local_size):
        kwds = dict(self._constant_kwds)
        fft_size = self._fft_size

        radix_array = get_radix_array(fft_size)
        if fft_size // radix_array[0] > max_local_size:
            radix_array = get_radix_array(fft_size, use_max_radix=True)

        threads_per_xform = fft_size // radix_array[0]
        local_size = max(64, threads_per_xform)
        if local_size > max_local_size:
            raise OutOfResourcesError
        xforms_per_workgroup = local_size // threads_per_xform
        workgroups_num = helpers.min_blocks(self._outer_batch, xforms_per_workgroup)

        lmem_size = get_local_memory_size(
            fft_size, radix_array, threads_per_xform, xforms_per_workgroup,
            kwds['local_mem_banks'], kwds['min_mem_coalesce_width'])

        if lmem_size * self._itemsize // 2 > self._local_mem_size:
            raise OutOfResourcesError

        kwds.update(dict(
            fft_size=fft_size, fft_size_real=self._fft_size_real, radix_arr=radix_array,
            lmem_size=lmem_size, threads_per_xform=threads_per_xform,
            xforms_per_workgroup=xforms_per_workgroup,
            outer_batch=self._outer_batch))

        return local_size * workgroups_num, local_size, kwds


class GlobalFFTKernel:
    """Generator for 'global' FFT kernel chain."""

    def __init__(self, dtype, device_params, outer_shape, fft_size, curr_size,
            fft_size_real, inner_shape, pass_num, reverse_direction):

        num_passes = len(get_global_radix_info(fft_size)[0])
        real_output_shape = (pass_num == num_passes - 1 and reverse_direction)

        self.name = 'fft_global'
        self.inplace_possible = (pass_num == num_passes - 1 and num_passes % 2 == 1)
        self.output_shape = (outer_shape +
            (fft_size_real if real_output_shape else fft_size,) + inner_shape)
        if fft_size != fft_size_real and pass_num == 0 and reverse_direction:
            self.kweights = get_kweights(fft_size_real, fft_size)
        else:
            self.kweights = None

        self._fft_size = fft_size
        self._curr_size = curr_size
        self._fft_size_real = fft_size_real
        self._local_mem_size = device_params.local_mem_size
        self._itemsize = dtype.itemsize
        self._inner_batch = helpers.product(inner_shape)
        self._outer_batch = helpers.product(outer_shape)
        self._pass_num = pass_num
        self._last_pass = (pass_num == num_passes - 1)

        self._constant_kwds = get_common_kwds(dtype, device_params)
        self._constant_kwds.update(dict(
            takes_kweights=(self.kweights is not None),
            input_slices=(len(outer_shape), 1, len(inner_shape)),
            output_slices=(len(outer_shape), 1, len(inner_shape)),
            pad_in=(fft_size != fft_size_real and pass_num == 0 and not reverse_direction),
            unpad_out=(fft_size != fft_size_real and self._last_pass and reverse_direction),
            reverse_direction=reverse_direction,
            normalize=self._last_pass))

    def prepare_for(self, max_local_size):
        kwds = dict(self._constant_kwds)

        radix_arr, radix1_arr, radix2_arr = get_global_radix_info(self._fft_size)

        radix = radix_arr[self._pass_num]
        radix1 = radix1_arr[self._pass_num]
        radix2 = radix2_arr[self._pass_num]

        stride_out = self._inner_batch * helpers.product(radix_arr[:self._pass_num])
        stride = stride_out * radix
        stride_in = stride_out * helpers.product(radix_arr[self._pass_num+1:])

        threads_per_xform = radix2

        coalesce_width = kwds['min_mem_coalesce_width']
        local_batch = max_local_size if radix2 == 1 else coalesce_width
        local_batch = min(local_batch, stride_in)
        local_size = min(local_batch * threads_per_xform, max_local_size)
        local_batch = local_size // threads_per_xform

        workgroups_num = helpers.min_blocks(stride_in, local_batch) * self._outer_batch

        if radix2 == 1:
            lmem_size = 0
        else:
            if stride_out == 1:
                lmem_size = (radix + 1) * local_batch
            else:
                lmem_size = local_size * radix1

        if lmem_size * self._itemsize // 2 > self._local_mem_size:
            raise OutOfResourcesError

        kwds.update(self._constant_kwds)
        kwds.update(dict(
            fft_size=self._fft_size, curr_size=self._curr_size, fft_size_real=self._fft_size_real,
            pass_num=self._pass_num,
            lmem_size=lmem_size, local_batch=local_batch, local_size=local_size,
            inner_batch=self._inner_batch,
            radix_arr=radix_arr, radix1_arr=radix1_arr, radix2_arr=radix2_arr,
            radix1=radix1, radix2=radix2, radix=radix,
            stride_in=stride_in, stride_out=stride_out, stride=stride,
            last_pass=self._last_pass))

        return workgroups_num * local_size, local_size, kwds

    @staticmethod
    def create_chain(dtype, device_params, outer_shape, fft_size, fft_size_real, inner_shape,
            reverse_direction):

        radix_arr, _, _ = get_global_radix_info(fft_size)

        curr_size = fft_size
        kernels = []
        for pass_num in range(len(radix_arr)):
            kernels.append(GlobalFFTKernel(
                dtype, device_params, outer_shape, fft_size,
                curr_size, fft_size_real, inner_shape, pass_num,
                reverse_direction))
            curr_size //= radix_arr[pass_num]

        return kernels


def get_fft_1d_kernels(dtype, device_params, outer_shape, fft_size, inner_shape,
        local_kernel_limit, reverse_direction=False, fft_size_real=None):
    """Create and compile kernels for one of the dimensions"""

    kernels = []

    if fft_size_real is None:
        fft_size_real = fft_size

    if (helpers.product(inner_shape) == 1 and fft_size // MAX_RADIX <= local_kernel_limit):
        kernels.append(LocalFFTKernel(
            dtype, device_params, outer_shape, fft_size, fft_size_real,
            inner_shape, reverse_direction))
    else:
        kernels.extend(GlobalFFTKernel.create_chain(
            dtype, device_params, outer_shape, fft_size, fft_size_real,
            inner_shape, reverse_direction))

    return kernels


def get_fft_kernels(input_shape, dtype, axes, device_params, local_kernel_limit):
    kernels = []

    # Starting from the most local transformation, for the sake of neatness.
    # Does not really matter.
    for axis in reversed(axes):
        outer_shape = input_shape[:axis]
        fft_size = input_shape[axis]
        inner_shape = input_shape[axis+1:]

        if fft_size == 1:
            continue

        bounding_size = helpers.bounding_power_of_2(fft_size)

        if bounding_size == fft_size:
            kernels.extend(get_fft_1d_kernels(
                dtype, device_params, outer_shape, fft_size,
                inner_shape, local_kernel_limit))
        else:
            # padding FFT for the chirp-z transform
            fft_size_padded = 2 * bounding_size
            args = (dtype, device_params, outer_shape, fft_size_padded,
                inner_shape, local_kernel_limit)

            new_kernels = []
            new_kernels.extend(get_fft_1d_kernels(
                *args, fft_size_real=fft_size))
            new_kernels.extend(get_fft_1d_kernels(
                *args, reverse_direction=True, fft_size_real=fft_size))

            # Since during pad-in or pad-out input and output blocks are no longer aligned,
            # these kernels lose their inplace_possible property
            new_kernels[0].inplace_possible = False
            new_kernels[-1].inplace_possible = False

            kernels.extend(new_kernels)

    return kernels


class LocalKernelFail(Exception):
    pass

class GlobalKernelFail(Exception):
    pass


class FFT(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Performs the Fast Fourier Transform.
    The interface is similar to ``numpy.fft.fftn``.
    The inverse transform is normalized so that ``IFFT(FFT(X)) = X``.

    :param arr_t: an array-like defining the problem array.
    :param axes: a tuple with axes over which to perform the transform.
        If not given, the transform is performed over all the axes.

    .. note::
        Current algorithm works most effectively with array dimensions being power of 2
        This mostly applies to the axes over which the transform is performed,
        beacuse otherwise the computation falls back to the Bluestein's algorithm,
        which effectively halves the performance.

    .. py:method:: compiled_signature(output:o, input:i, inverse:s)

        ``output`` and ``input`` may be the same array.

        ..
            The only case where it would matter is when we only have one kernel in the plan,
            and it has ``inplace_possible==False``.
            Local kernels are always inplace, so it must be a global kernel.
            But if there is only one global kernel in the list, it will have its
            ``inplace_possible`` set to ``True`` (see the condition in ``GlobalFFTKernel``).
            Therefore our FFT is always guaranteed to be inplace.

        :param output: an array with the attributes of ``arr_t``.
        :param input: an array with the attributes of ``arr_t``.
        :param inverse: a scalar value castable to integer.
            If ``1``, ``output`` contains the forward FFT of ``input``,
            if ``0`` the inverse one.
    """

    def __init__(self, arr_t, axes=None):

        if not dtypes.is_complex(arr_t.dtype):
            raise ValueError("FFT computation requires array of a complex dtype")

        Computation.__init__(self, [
            Parameter('output', Annotation(arr_t, 'o')),
            Parameter('input', Annotation(arr_t, 'i')),
            Parameter('inverse', Annotation(numpy.int32), default=0)])

        if axes is None:
            axes = tuple(range(len(arr_t.shape)))
        else:
            axes = tuple(axes)
        self._axes = axes

    def _build_trivial_plan(self, plan_factory, output, input_):
        # Trivial problem. Need to add a dummy kernel
        # because we still have to run transformations.

        plan = plan_factory()

        copy_trf = copy(input_, out_arr_t=output)
        copy_comp = PureParallel.from_trf(copy_trf, copy_trf.input)
        plan.computation_call(copy_comp, output, input_)

        return plan

    def _build_limited_plan(self, plan_factory, device_params, local_kernel_limit,
            output, input_, inverse):

        plan = plan_factory()
        kernels = get_fft_kernels(
            input_.shape, input_.dtype, self._axes, device_params, local_kernel_limit)

        mem_out = None
        for i, kernel in enumerate(kernels):

            mem_in = input_ if i == 0 else mem_out
            if i == len(kernels) - 1:
                mem_out = output
            elif kernel.inplace_possible and mem_in is not input_:
                mem_out = mem_in
            else:
                mem_out = plan.temp_array(kernel.output_shape, output.dtype)

            if kernel.kweights is not None:
                kweights = plan.persistent_array(kernel.kweights.astype(output.dtype))
                kweights_arg = [kweights]
            else:
                kweights_arg = []

            argnames = [mem_out, mem_in] + kweights_arg + [inverse]

            # Try to find local size for each of the kernels
            local_size = device_params.max_work_group_size
            while local_size >= 1:
                try:
                    gsize, lsize, kwds = kernel.prepare_for(local_size)
                    plan.kernel_call(
                        TEMPLATE.get_def(kernel.name), argnames,
                        global_size=gsize, local_size=lsize, render_kwds=kwds)
                except OutOfResourcesError:
                    if isinstance(kernel, GlobalFFTKernel):
                        local_size //= 2
                        continue
                    else:
                        raise LocalKernelFail
                break
            else:
                raise GlobalKernelFail

        return plan

    def _build_plan(self, plan_factory, device_params, output, input_, inverse):

        if helpers.product([input_.shape[i] for i in self._axes]) == 1:
            return self._build_trivial_plan(plan_factory, output, input_)

        # While resource consumption of GlobalFFTKernel can be made lower by passing
        # lower value to prepare_for(), LocalFFTKernel may have to be split into several kernels.
        # Therefore, if GlobalFFTKernel.prepare_for() raises OutOfResourcesError,
        # we just call prepare_for() with lower limit, but if LocalFFTKernel.prepare_for()
        # does that, we have to recreate the whole chain.
        local_kernel_limit = device_params.max_work_group_size

        while local_kernel_limit >= 1:
            try:
                plan = self._build_limited_plan(
                    plan_factory, device_params, local_kernel_limit, output, input_, inverse)
            except LocalKernelFail:
            # One of LocalFFTKernels was out of resources.
            # Reduce the limit and try to create operations from scratch again.
                local_kernel_limit //= 2
                continue
            except GlobalKernelFail:
                raise ValueError(
                    "Could not find suitable call parameters for one of the global kernels")

            return plan

        raise ValueError("Could not find suitable call parameters for one of the local kernels")



"""
This module contains functions for API discovery.
"""


def cuda_id():
    """Returns the identifier of the ``PyCUDA``-based API."""
    return 'cuda'

def ocl_id():
    """Returns the identifier of the ``PyOpenCL``-based API."""
    return 'ocl'

def api_ids():
    """
    Returns a list of identifiers for all known
    (not necessarily available for the current system) APIs.
    """
    return [ocl_id(), cuda_id()]


def supports_api(api_id):
    """
    Returns ``True`` if given API is supported.
    """
    try:
        get_api(api_id)
    except ImportError:
        return False

    return True


def supported_api_ids():
    """
    Returns a list of identifiers of supported APIs.
    """
    return [api_id for api_id in api_ids() if supports_api(api_id)]


def get_api(api_id):
    """
    Returns an API module with the generalized interface :py:mod:`reikna.cluda.api`
    for the given identifier.
    """
    if api_id == cuda_id():
        import reikna.cluda.cuda
        return reikna.cluda.cuda
    elif api_id == ocl_id():
        import reikna.cluda.ocl
        return reikna.cluda.ocl
    else:
        raise ValueError("Unrecognized API: " + str(api_id))


def cuda_api():
    """
    Returns the ``PyCUDA``-based API module.
    """
    return get_api(cuda_id())


def ocl_api():
    """
    Returns the ``PyOpenCL``-based API module.
    """
    return get_api(ocl_id())


def any_api():
    """
    Returns one of the API modules supported by the system or raises an ``Exception``
    if there are not any.
    """
    api_ids = supported_api_ids()
    if len(api_ids) == 0:
        raise Exception("No supported GPGPU APIs found")
    return get_api(api_ids[0])



import funcsigs # backport of inspect.signature() and related objects for Py2
import numpy

import reikna.helpers as helpers
import reikna.cluda.dtypes as dtypes
from reikna.helpers import wrap_in_tuple, product


class Type:
    """
    Represents an array or, as a degenerate case, scalar type of a computation parameter.

    .. py:attribute:: shape

        A tuple of integers.
        Scalars are represented by an empty tuple.

    .. py:attribute:: dtype

        A ``numpy.dtype`` instance.

    .. py:attribute:: ctype

        A string with the name of C type corresponding to :py:attr:`dtype`,
        or a module if it is a struct type.

    .. py:attribute:: strides

        Tuple of bytes to step in each dimension when traversing an array.
    """

    def __init__(self, dtype, shape=None, strides=None):
        self.shape = tuple() if shape is None else wrap_in_tuple(shape)
        self.size = product(self.shape)
        self.dtype = dtypes.normalize_type(dtype)
        self.ctype = dtypes.ctype_module(self.dtype)
        if strides is None:
            self.strides = tuple([
                self.dtype.itemsize * product(self.shape[i+1:]) for i in range(len(self.shape))])
        else:
            self.strides = strides
        self._cast = dtypes.cast(self.dtype)

    def __eq__(self, other):
        return (self.shape == other.shape and self.dtype == other.dtype
            and self.strides == other.strides)

    def __ne__(self, other):
        return not (self == other)

    def compatible_with(self, other):
        if self.dtype != other.dtype:
            return False

        common_shape_len = min(len(self.shape), len(other.shape))
        if self.shape[-common_shape_len:] != other.shape[-common_shape_len:]:
            return False
        if self.strides[-common_shape_len:] != other.strides[-common_shape_len:]:
            return False
        if helpers.product(self.shape[:-common_shape_len]) != 1:
            return False
        if helpers.product(other.shape[:-common_shape_len]) != 1:
            return False

        return True

    @classmethod
    def from_value(cls, val):
        """
        Creates a :py:class:`Type` object corresponding to the given value.
        """
        if isinstance(val, Type):
            # Creating a new object, because ``val`` may be some derivative of Type,
            # used as a syntactic sugar, and we do not want it to confuse us later.
            return cls(val.dtype, shape=val.shape, strides=val.strides)
        elif numpy.issctype(val):
            return cls(val)
        elif hasattr(val, 'dtype') and hasattr(val, 'shape'):
            strides = val.strides if hasattr(val, 'strides') else None
            return cls(val.dtype, shape=val.shape, strides=strides)
        else:
            return cls(dtypes.detect_type(val))

    def __call__(self, val):
        """
        Casts the given value to this type.
        """
        return self._cast(val)

    def __repr__(self):
        if len(self.shape) > 0:
            return "Type({dtype}, shape={shape}, strides={strides})".format(
                dtype=self.dtype, shape=self.shape, strides=self.strides)
        else:
            return "Type({dtype})".format(dtype=self.dtype)

    def __process_modules__(self, process):
        tp = Type(self.dtype, shape=self.shape, strides=self.strides)
        tp.ctype = process(tp.ctype)
        return tp


class Annotation:
    """
    Computation parameter annotation,
    in the same sense as it is used for functions in the standard library.

    :param type_: a :py:class:`~reikna.core.Type` object.
    :param role: any of ``'i'`` (input), ``'o'`` (output),
        ``'io'`` (input/output), ``'s'`` (scalar).
        Defaults to ``'s'`` for scalars and ``'io'`` for arrays.
    """

    def __init__(self, type_, role=None):
        self.type = Type.from_value(type_)

        if role is None:
            role = 's' if len(self.type.shape) == 0 else 'io'

        assert role in ('i', 'o', 'io', 's')
        self.role = role
        if role == 's':
            self.array = False
            self.input = False
            self.output = False
        else:
            self.array = True
            self.input = 'i' in role
            self.output = 'o' in role

    def __eq__(self, other):
        return self.type == other.type and self.role == other.role

    def can_be_argument_for(self, annotation):
        if not self.type.compatible_with(annotation.type):
            return False

        if self.role == annotation.role:
            return True

        if self.role == 'io' and annotation.array:
            return True

        return False

    def __repr__(self):
        if self.array:
            return "Annotation({type_}, role={role})".format(
                type_=self.type, role=repr(self.role))
        else:
            return "Annotation({dtype})".format(dtype=self.type.dtype)

    def __process_modules__(self, process):
        ann = Annotation(self.type, role=self.role)
        ann.type = process(ann.type)
        return ann


class Parameter(funcsigs.Parameter):
    """
    Computation parameter,
    in the same sense as it is used for functions in the standard library.
    In its terms, all computation parameters have kind ``POSITIONAL_OR_KEYWORD``.

    :param name: parameter name.
    :param annotation: an :py:class:`~reikna.core.Annotation` object.
    :param default: default value for the parameter, can only be specified for scalars.
    """

    def __init__(self, name, annotation, default=funcsigs.Parameter.empty):

        if default is not funcsigs.Parameter.empty:
            if annotation.array:
                raise ValueError("Array parameters cannot have default values")
            default = annotation.type(default)

        # HACK: Parameter constructor is not documented.
        # But I need to create these objects somehow.
        funcsigs.Parameter.__init__(
            self, name, annotation=annotation,
            kind=funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
            default=default)

    def rename(self, new_name):
        """
        Creates a new :py:class:`Parameter` object with the new name
        and the same annotation and default value.
        """
        return Parameter(new_name, self.annotation, default=self.default)

    def __eq__(self, other):
        return (self.name == other.name and self.annotation == other.annotation
            and self.default == other.default)

    def __process_modules__(self, process):
        return Parameter(self.name, process(self.annotation), default=self.default)


class Signature(funcsigs.Signature):
    """
    Computation signature,
    in the same sense as it is used for functions in the standard library.

    :param parameters: a list of :py:class:`~reikna.core.Parameter` objects.

    .. py:attribute:: parameters

        An ``OrderedDict`` with :py:class:`~reikna.core.Parameter` objects indexed by their names.
    """

    def __init__(self, parameters):
        # HACK: Signature constructor is not documented.
        # But I need to create these objects somehow.
        funcsigs.Signature.__init__(self, parameters)

    def bind_with_defaults(self, args, kwds, cast=False):
        """
        Binds passed positional and keyword arguments to parameters in the signature and
        returns the resulting ``BoundArguments`` object.
        """
        bound_args = self.bind(*args, **kwds)
        for param in self.parameters.values():
            if param.name not in bound_args.arguments:
                bound_args.arguments[param.name] = param.default
            elif cast:
                if not param.annotation.array:
                    bound_args.arguments[param.name] = \
                        param.annotation.type(bound_args.arguments[param.name])
        return bound_args



import sys
import itertools

import pycuda.gpuarray as gpuarray
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
from pycuda.tools import DeviceData

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
from reikna.helpers import factors, wrap_in_tuple, product
import reikna.cluda.api as api_base


cuda.init()


def get_id():
    return cluda.cuda_id()


def get_platforms():
    # For CUDA, there's only one platform
    return [Platform()]


class Platform:
    """
    Mimics pyopencl.Platform
    """

    name = "nVidia CUDA"
    vendor = "nVidia"
    version = ".".join(str(x) for x in cuda.get_version())

    def get_devices(self):
        return [Device(num) for num in range(cuda.Device.count())]

    def __str__(self):
        return self.name + " " + self.version


class Device(cuda.Device):

    def __init__(self, device_num):
        cuda.Device.__init__(self, device_num)
        self.name = self.name()


class Buffer:
    """
    Mimics pyopencl.Buffer
    """

    def __init__(self, size):
        self._buffer = cuda.mem_alloc(size)
        self.size = size

    def __int__(self):
        return int(self._buffer)

    def __long__(self):
        return long(self._buffer)

    def __del__(self):
        self._buffer.free()


class Array(gpuarray.GPUArray):
    """
    A superclass of PyCUDA ``GPUArray``, with some additional functionality.
    """
    def __init__(self, thr, *args, **kwds):
        gpuarray.GPUArray.__init__(self, *args, **kwds)
        self.thread = thr


class Thread(api_base.Thread):

    api = sys.modules[__name__]

    def __init__(self, *args, **kwds):
        api_base.Thread.__init__(self, *args, **kwds)
        self._active = True

    def _process_cqd(self, cqd):
        if isinstance(cqd, cuda.Device):
            context = cqd.make_context()
            stream = cuda.Stream()
            return context, stream, cqd, True
        elif isinstance(cqd, cuda.Context) or cqd is None:
            return cqd, cuda.Stream(), cqd.get_device(), False
        elif isinstance(cqd, cuda.Stream):
            # There's no function in PyCUDA to get the current context,
            # but we do not really need it anyway.
            return None, cqd, cuda.Context.get_device(), False
        else:
            return ValueError("The value provided is not Device, Context or Stream")

    def allocate(self, size):
        return Buffer(size)

    def array(self, shape, dtype, strides=None, allocator=None):
        # In PyCUDA, the default allocator is not None, but a default alloc object
        kwds = {}
        if strides is not None:
            kwds['strides'] = strides
        if allocator is not None:
            kwds['allocator'] = allocator
        return Array(self, shape, dtype, **kwds)

    def _copy_array(self, dest, src):
        dest.set_async(src, stream=self._queue)

    def from_device(self, arr, dest=None, async=False):
        if async:
            arr_cpu = arr.get_async(ary=dest, stream=self._queue)
        else:
            arr_cpu = arr.get(ary=dest)

        if dest is None:
            return arr_cpu

    def _copy_array_buffer(self, dest, src, nbytes, src_offset=0, dest_offset=0):
        cuda.memcpy_dtod_async(
            int(dest.gpudata) + dest_offset,
            int(src.gpudata) + src_offset,
            nbytes, stream=self._queue)

    def synchronize(self):
        self._queue.synchronize()

    def _compile(self, src, fast_math=False):
        options = ['-use_fast_math'] if fast_math else []
        return SourceModule(src, no_extern_c=True, options=options)

    def _cuda_push(self):
        assert not self._active
        self._context.push()
        self._active = True

    def _cuda_pop(self):
        assert self._active
        cuda.Context.pop()
        self._active = False

    def _release_specific(self):
        # If we own the context, it is our responsibility to pop() it
        if self._owns_context and self._active:
            cuda.Context.pop()
        if self._owns_context:
            self._context.detach()

    def __del__(self):
        self.release()


class DeviceParameters:

    def __init__(self, device):

        self._device = device
        self.max_work_group_size = device.max_threads_per_block
        self.max_work_item_sizes = [
            device.max_block_dim_x,
            device.max_block_dim_y,
            device.max_block_dim_z]

        self.max_num_groups = [
            device.max_grid_dim_x,
            device.max_grid_dim_y,
            device.max_grid_dim_z]

        # there is no corresponding constant in the API at the moment
        self.local_mem_banks = 16 if device.compute_capability()[0] < 2 else 32

        self.warp_size = device.warp_size

        devdata = DeviceData(device)
        self.min_mem_coalesce_width = dict(
            ((size,devdata.align_words(word_size=size)) for size in [4, 8, 16]))
        self.local_mem_size = device.max_shared_memory_per_block

    def supports_dtype(self, dtype):
        if dtypes.is_double(dtype):
            major, minor = self._device.compute_capability()
            return (major == 1 and minor == 3) or major >= 2
        else:
            return True


def find_local_size(global_size, max_work_item_sizes, max_work_group_size):
    """
    Mimics the OpenCL local size finding algorithm.
    Returns the tuple of the same length as ``global_size``, with every element
    being a factor of the corresponding element of ``global_size``.
    Neither of the elements of ``local_size`` are greater then the corresponding element
    of ``max_work_item_sizes``, and their product is not greater than ``max_work_group_size``.
    """
    if len(global_size) == 0:
        return tuple()

    if max_work_group_size == 1:
        return (1,) * len(global_size)

    gs_factors = factors(global_size[0], limit=min(max_work_item_sizes[0], max_work_group_size))
    local_size_1d, _ = gs_factors[-1]
    remainder = find_local_size(
        global_size[1:], max_work_item_sizes[1:], max_work_group_size // local_size_1d)

    return (local_size_1d,) + remainder


class Kernel(api_base.Kernel):

    def _get_kernel(self, program, name):
        return program.get_function(name)

    def _fill_attributes(self):
        self.max_work_group_size = self._kernel.get_attribute(
            cuda.function_attribute.MAX_THREADS_PER_BLOCK)

    def prepare(self, global_size, local_size=None, local_mem=0):
        global_size = wrap_in_tuple(global_size)
        self._local_mem = local_mem

        max_dims = self._thr.device_params.max_work_item_sizes
        if len(global_size) > len(max_dims):
            raise ValueError("Global size has too many dimensions")

        if local_size is not None:
            local_size = wrap_in_tuple(local_size)
            if len(local_size) != len(global_size):
                raise ValueError("Global/local work sizes have differing dimensions")
        else:
            local_size = find_local_size(global_size, max_dims, self.max_work_group_size)

        grid = []
        for gs, ls in zip(global_size, local_size):
            if gs % ls != 0:
                raise ValueError("Global sizes must be multiples of corresponding local sizes")
            grid.append(gs // ls)

        # append missing dimensions, otherwise PyCUDA will complain
        self._local_size = local_size + (1,) * (3 - len(grid))
        self._grid = tuple(grid) + (1,) * (3 - len(grid))

    def prepared_call(self, *args):
        self._kernel(*args, grid=self._grid, block=self._local_size,
            stream=self._thr._queue, shared=self._local_mem)
